from pydantic import BaseModel, NonNegativeInt

from rethink import const


class VerifyCaptchaRequest(BaseModel):
    token: str
    codeStr: str
    language: str = const.Language.EN.value
    requestId: str = ""


class VerifyCaptchaResponse(BaseModel):
    code: NonNegativeInt
    message: str
    requestId: str




from fastapi import APIRouter
from fastapi.responses import StreamingResponse

from rethink.controllers import schemas
from rethink.controllers.verification import v_captcha
from rethink.routes.utils import measure_time_spend

router = APIRouter(
    prefix="/api",
    tags=["verification"],
    responses={404: {"description": "Not found"}},
)


@router.get(
    path="/captcha/img",
)
@measure_time_spend
async def get_captcha_img() -> StreamingResponse:
    return v_captcha.get_captcha_img()


@router.post(
    path="/captcha/img",
    response_model=schemas.verification.VerifyCaptchaResponse,
)
@measure_time_spend
async def verify_captcha(
        req: schemas.verification.VerifyCaptchaRequest,
) -> schemas.verification.VerifyCaptchaResponse:
    return v_captcha.verify_captcha_resp(req=req)



from typing import List, Optional

from pydantic import BaseModel, NonNegativeInt, Field

from ..utils import datetime2str


class NodeData(BaseModel):
    class LinkedNode(BaseModel):
        id: str
        title: str
        md: str
        snippet: str
        type: NonNegativeInt
        disabled: bool
        inTrash: bool
        createdAt: str
        modifiedAt: str

    id: str
    md: str
    title: str
    snippet: str
    type: NonNegativeInt
    disabled: bool
    createdAt: str
    modifiedAt: str
    fromNodes: List[LinkedNode] = Field(default_factory=list)
    toNodes: List[LinkedNode] = Field(default_factory=list)


class PutRequest(BaseModel):
    md: str
    type: NonNegativeInt
    requestId: str = ""
    fromNid: str = ""


class PutResponse(BaseModel):
    code: NonNegativeInt
    message: str
    requestId: str
    node: Optional[NodeData]


class GetResponse(BaseModel):
    code: NonNegativeInt
    message: str
    requestId: str
    node: Optional[NodeData]


class UpdateRequest(BaseModel):
    nid: str
    md: str
    requestId: str = ""


class NodesInfoResponse(BaseModel):
    class Data(BaseModel):
        class NodeInfo(BaseModel):
            id: str
            title: str
            snippet: str
            type: int
            createdAt: str
            modifiedAt: str

        nodes: List[NodeInfo]
        total: NonNegativeInt

    code: NonNegativeInt
    message: str
    requestId: str
    data: Data


class RestoreFromTrashRequest(BaseModel):
    requestId: str
    nid: str


class GetFromTrashResponse(BaseModel):
    code: NonNegativeInt
    message: str
    requestId: str
    data: NodesInfoResponse.Data


def parse_nodes_info(nodes, total):
    data = NodesInfoResponse.Data(
        nodes=[NodesInfoResponse.Data.NodeInfo(
            id=n["id"],
            title=n["title"],
            snippet=n["snippet"],
            type=n["type"],
            createdAt=datetime2str(n["_id"].generation_time),
            modifiedAt=datetime2str(n["modifiedAt"]),
        ) for n in nodes],
        total=total,
    )
    return data


class BatchNodeIdsRequest(BaseModel):
    requestId: str
    nids: List[str]



from typing import List, Tuple, Sequence

from rethink import const, config
from .. import tps
from ..database import COLL


def user_node(
        uid: str,
        query: str = "",
        sort_key: str = "createdAt",
        sort_order: int = -1,
        page: int = 0,
        page_size: int = 0,
        nid_exclude: Sequence[str] = None,
) -> Tuple[List[tps.Node], int]:
    condition = {
        "uid": uid,
        "disabled": False,
        "inTrash": False,
    }
    if nid_exclude is not None and len(nid_exclude) > 0:
        condition["id"] = {"$nin": nid_exclude}

    total = COLL.nodes.count_documents(condition)

    query = query.strip().lower()

    # on remote mongodb
    if query != "" and not config.is_local_db():
        condition["$or"] = [
            {"searchKeys": {"$regex": query}},
            {"md": {"$regex": query}},
        ]

    docs = COLL.nodes.find(condition)

    if sort_key != "":
        if sort_key == "createdAt":
            sort_key = "_id"
        elif sort_key == "similarity":
            sort_key = "_id"  # TODO: sort by similarity
        docs = docs.sort([(sort_key, sort_order), ("_id", -1)])

    if config.is_local_db() and query != "":
        docs = filter(lambda d: query in d["searchKeys"] or query in d["md"], docs)
        docs = list(docs)
        if page_size > 0:
            docs = docs[page * page_size: (page + 1) * page_size]
        return docs, total

    if page_size > 0:
        docs = docs.skip(page * page_size).limit(page_size)
    return list(docs), total


def add_recent_cursor_search(
        uid: str,
        nid: str,
        to_nid: str,
) -> const.Code:
    # add selected node to recentCursorSearchSelectedNIds
    user_c = {"id": uid, "disabled": False}
    node_c = {"uid": uid, "id": {"$in": [nid, to_nid]}}

    # try finding user
    u = COLL.users.find_one(user_c)
    if u is None:
        return const.Code.ACCOUNT_OR_PASSWORD_ERROR

    # try finding node
    ns = list(COLL.nodes.find(node_c))
    if len(ns) != 2:
        return const.Code.NODE_NOT_EXIST

    rns = u["lastState"]["recentCursorSearchSelectedNIds"]
    if to_nid in rns:
        rns.remove(to_nid)
    rns.insert(0, to_nid)
    if len(rns) > 10:
        rns = rns[:10]

    # add to recentCursorSearchSelectedNIds
    res = COLL.users.update_one(
        {"id": uid},
        {"$set": {"lastState.recentCursorSearchSelectedNIds": rns}}
    )
    if res.matched_count != 1:
        return const.Code.OPERATION_FAILED

    return const.Code.OK


def get_recent_search(uid: str) -> List[str]:
    doc = COLL.users.find_one({"id": uid})
    if doc is None:
        return []
    return doc["lastState"]["recentSearch"]


def put_recent_search(uid: str, query: str) -> const.Code:
    doc = COLL.users.find_one({"id": uid})
    if doc is None:
        return const.Code.ACCOUNT_OR_PASSWORD_ERROR
    rns = doc["lastState"]["recentSearch"]
    try:
        rns.remove(query)
    except ValueError:
        pass
    rns.insert(0, query)
    if len(rns) > 20:
        rns = rns[:20]
    _ = COLL.users.update_one(
        {"id": uid},
        {"$set": {"lastState.recentSearch": rns}}
    )
    return const.Code.OK



import datetime
from typing import Optional, Tuple

from bson import ObjectId
from bson.tz_util import utc

from rethink import const, config
from . import tps, utils
from .database import COLL


def add(
        account: str,
        source: int,
        email: str,
        hashed: str,
        nickname: str,
        avatar: str,
        language: str,
) -> Tuple[str, const.Code]:
    if COLL.users.find_one({"account": account, "source": source}) is not None:
        return "", const.Code.EMAIL_OCCUPIED
    oid = ObjectId()
    # assert language in const.Language
    if not const.Language.is_valid(language):
        language = const.Language.EN.value
    data: tps.UserMeta = {
        "_id": oid,
        "id": utils.short_uuid(),
        "account": account,
        "source": source,
        "email": email,
        "hashed": hashed,
        "avatar": str(avatar),
        "disabled": False,
        "nickname": nickname,
        "modifiedAt": oid.generation_time,
        "nodeIds": [],
        "language": language,
        "usedSpace": 0,
        "type": const.USER_TYPE.NORMAL.id,
        "lastState": {
            "recentCursorSearchSelectedNIds": [],
            "recentSearch": [],
            "nodeDisplayMethod": const.NodeDisplayMethod.CARD.value,
            "nodeDisplaySortKey": "modifiedAt"
        }
    }
    res = COLL.users.insert_one(data)
    if not res.acknowledged:
        return "", const.Code.OPERATION_FAILED
    un_data: tps.UserNodeIds = {
        "_id": data["_id"],
        "id": data["id"],
        "nodeIds": [],
    }
    return data["id"], const.Code.OK


def update(
        uid: str,
        email: str = "",
        hashed: str = "",
        nickname: str = "",
        avatar: str = "",
        language: str = "",
        node_display_method: int = -1,
        node_display_sort_key: str = "",
) -> Tuple[Optional[tps.UserMeta], const.Code]:
    u, code = get(uid=uid)
    if code != const.Code.OK:
        return None, code
    email = email.strip()
    if email not in ["", u["email"]] and COLL.users.find_one(
            {"account": email, "source": const.UserSource.EMAIL.value}
    ) is not None:
        return None, const.Code.EMAIL_OCCUPIED

    new_data = {"modifiedAt": datetime.datetime.now(tz=utc), }
    if email != "" and email != u["email"]:
        new_data["email"] = email

    hashed = hashed.strip()
    if hashed != "" and hashed != u["hashed"]:
        new_data["hashed"] = hashed
    nickname = nickname.strip()
    if nickname != "" and nickname != u["nickname"]:
        new_data["nickname"] = nickname
    avatar = str(avatar).strip()
    if avatar != "" and avatar != u["avatar"]:
        new_data["avatar"] = avatar
    language = language.strip()
    if language != "" and language != u["language"]:
        if not const.Language.is_valid(language):
            return None, const.Code.INVALID_LANGUAGE
        new_data["language"] = language

    if node_display_method != u["lastState"]["nodeDisplayMethod"] and node_display_method >= 0:
        if node_display_method >= len(const.NodeDisplayMethod):
            return None, const.Code.INVALID_NODE_DISPLAY_METHOD
        new_data["lastState.nodeDisplayMethod"] = node_display_method
    if node_display_sort_key != "" and node_display_sort_key != u["lastState"]["nodeDisplaySortKey"]:
        if node_display_sort_key not in ["modifiedAt", "createdAt", "title"]:
            return None, const.Code.INVALID_NODE_DISPLAY_SORT_KEY
        new_data["lastState.nodeDisplaySortKey"] = node_display_sort_key

    res = COLL.users.update_one(
        {"id": uid},
        {"$set": new_data},
    )
    if res.modified_count != 1:
        return None, const.Code.OPERATION_FAILED
    return get(uid=uid)


def delete(uid: str) -> const.Code:
    res = COLL.users.delete_one({"id": uid})
    return const.Code.OK if res.deleted_count == 1 else const.Code.OPERATION_FAILED


def disable(uid: str) -> const.Code:
    res = COLL.users.update_one(
        {"id": uid},
        {"$set": {"disabled": True}}
    )
    return const.Code.OK if res.modified_count == 1 else const.Code.OPERATION_FAILED


def enable(uid: str) -> const.Code:
    res = COLL.users.update_one(
        {"id": uid},
        {"$set": {"disabled": False}}
    )
    return const.Code.OK if res.modified_count == 1 else const.Code.OPERATION_FAILED


def get_by_email(email: str) -> Tuple[Optional[tps.UserMeta], const.Code]:
    if config.get_settings().ONE_USER:
        source = const.UserSource.LOCAL.value
    else:
        source = const.UserSource.EMAIL.value
    return get_account(account=email, source=source)


def get_account(account: str, source: int) -> Tuple[Optional[tps.UserMeta], const.Code]:
    u = COLL.users.find_one({"source": source, "account": account, "disabled": False})
    if u is None:
        return None, const.Code.ACCOUNT_OR_PASSWORD_ERROR
    return u, const.Code.OK


def get(uid: str) -> Tuple[Optional[tps.UserMeta], const.Code]:
    u = COLL.users.find_one({"id": uid, "disabled": False})
    if u is None:
        return None, const.Code.ACCOUNT_OR_PASSWORD_ERROR
    if u["usedSpace"] < 0:
        # reset usedSpace to 0
        COLL.users.update_one(
            {"id": uid},
            {"$set": {"usedSpace": 0}}
        )
        u["usedSpace"] = 0

    return u, const.Code.OK


def get_hash_by_uid(uid: str) -> Optional[str]:
    u = COLL.users.find_one({"id": uid, "disabled": False})
    if u is None:
        return None
    return u["hashed"]


def is_exist(uid: str) -> bool:
    try:
        COLL.users.find({"id": uid, "disabled": False}, limit=1).next()
    except StopIteration:
        return False
    return True


def update_used_space(uid: str, delta: int) -> const.Code:
    if not is_exist(uid=uid):
        return const.Code.ACCOUNT_OR_PASSWORD_ERROR
    res = COLL.users.update_one(
        {"id": uid},
        {"$inc": {"usedSpace": delta}}
    )
    return const.Code.OK if res.modified_count == 1 else const.Code.OPERATION_FAILED


def user_space_not_enough(uid: str = None, u: tps.UserMeta = None) -> bool:
    if uid is None and u is None:
        raise ValueError("uid and u cannot be None at the same time")
    if config.is_local_db():
        return False
    if uid is not None:
        u, code = get(uid=uid)
        if code != const.Code.OK:
            return True
    return u["usedSpace"] > const.USER_TYPE.id2config(u["type"]).max_store_space



import time
from functools import wraps

from rethink.logger import logger


def measure_time_spend(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        t0 = time.perf_counter()
        req_id = ""
        req_s = ""
        if "req" in kwargs:
            req = kwargs["req"]
            try:
                req_id = req.requestId
            except AttributeError:
                req_id = ""
            req_s = str(req)
        if "rid" in kwargs:
            req_id = kwargs["rid"]
        uid = ""
        code = ""
        if "token_decode" in kwargs:
            uid = kwargs["token_decode"].uid
            code = kwargs["token_decode"].code
        req_s = req_s[:200] + "..." if len(req_s) > 200 else req_s
        logger.info(f"REQ: reqId='{req_id}' | uid='{uid}' | api='{func.__name__}' | code='{code}' | req='{req_s}'")

        resp = await func(*args, **kwargs)
        t1 = time.perf_counter()
        try:
            req_id = resp.requestId
        except AttributeError:
            req_id = ""
        logger.info(f"RESP: reqId='{req_id}' | uid='{uid}' | api='{func.__name__}' | spend={t1 - t0:.4f}s")
        return resp

    return wrapper



from fastapi.responses import StreamingResponse

from rethink import const
from rethink.controllers import schemas
from rethink.models.verification import random_captcha, verify_captcha


def get_captcha_img():
    token, data = random_captcha(length=4, sound=False)
    return StreamingResponse(
        data["img"],
        headers={
            "X-Captcha-Token": token
        },
        media_type="image/png",
    )


def verify_captcha_resp(
        req: schemas.verification.VerifyCaptchaRequest
) -> schemas.verification.VerifyCaptchaResponse:
    code = verify_captcha(token=req.token, code_str=req.codeStr)
    _msg = const.CODE_MESSAGES[code]
    if req.language == const.Language.ZH.value:
        msg = _msg.zh
    elif req.language == const.Language.EN.value:
        msg = _msg.en
    else:
        msg = _msg.en
    return schemas.verification.VerifyCaptchaResponse(
        code=code.value,
        message=msg,
        requestId=req.requestId,
    )



import unittest
from textwrap import dedent

from rethink.models import utils


class UtilsTest(unittest.TestCase):
    def test_short_uuid(self):
        for _ in range(10):
            uid = utils.short_uuid()
            self.assertEqual(22, len(uid))

    def test_md2txt(self):
        text = utils.md2txt(md="# 123\n## 456\n### 789\n")
        self.assertEqual("123\n456\n789", text)

    def test_pinyin(self):
        res = utils.txt2search_keys('中心English')
        self.assertEqual({'zhongxinenglish', '中心english', 'ㄓㄨㄥㄒㄧㄣenglish'}, set(res.split(" ")))

    def test_change_link_title(self):
        md = dedent("""\
            # 123
            ddd qwd [@我是谁](/n/sqaaq) 345
            [@ssw](/n/weee)
            """)

        new_md = utils.change_link_title(md, nid="sqaaq", new_title="哇塞")
        self.assertEqual(dedent(f"""\
            # 123
            ddd qwd [@哇塞](/n/sqaaq) 345
            [@ssw](/n/weee)
            """), new_md)

        new_md = utils.change_link_title(md, nid="weee", new_title="哇塞")
        self.assertEqual(dedent(f"""\
            # 123
            ddd qwd [@我是谁](/n/sqaaq) 345
            [@哇塞](/n/weee)
            """), new_md)

        md = dedent("""\
            # 123
            ddd qwd [@我是谁](/n/sqaaq) 345
            [@我是谁](/n/sqaaq)
            """)
        new_md = utils.change_link_title(md, nid="sqaaq", new_title="哇塞")
        self.assertEqual(dedent(f"""\
            # 123
            ddd qwd [@哇塞](/n/sqaaq) 345
            [@哇塞](/n/sqaaq)
            """), new_md)

        md = dedent("""\
            # 123
            ddd qwd 345
            [@我](/n/1) [@我](/n/2)
            """)
        new_md = utils.change_link_title(md, nid="2", new_title="哇塞")
        self.assertEqual(dedent(f"""\
            # 123
            ddd qwd 345
            [@我](/n/1) [@哇塞](/n/2)
            """), new_md)

    def test_contain_only_link(self):
        for md, res in [
            ("http://rethink.run", "http://rethink.run"),
            ("[123](/n/123)", ""),
            ("few", ""),
            ("c https://rethink.run", ""),
            (" https://rethink.run  ", "https://rethink.run"),
            ("https://rethink.run  ", "https://rethink.run"),
            ("https://rethink.run", "https://rethink.run"),
            ("https://rethink.run wwq", ""),

        ]:
            self.assertEqual(res, utils.contain_only_http_link(md))


class TestAsync(unittest.IsolatedAsyncioTestCase):
    async def test_get_title_description_from_link(self):
        for url, res in [
            ("https://github.com/MorvanZhou/rethink", True),
            # ("https://zhuanlan.zhihu.com/p/610939462?utm_id=0", True),
            ("https://waqwe12f2f2fa.fffffffff", False),
            ("https://baidu.com", True),
            ("https://rethink.run", True),
            ("https://rethink.run/about", True),
            ("https://baidu.com/wqwqqqqq", False),
            ("https://mp.weixin.qq.com/s/jbB0GXbjHpFR8m1-6TSASw", False),
        ]:
            title, desc = await utils.get_title_description_from_link(url)
            if res:
                self.assertNotEqual(title, "", msg=f"{url} {title}")
                self.assertNotEqual(desc, "", msg=f"{url} {desc}")
            else:
                self.assertEqual(title, "", msg=f"{url} {title}")
                self.assertEqual(desc, "", msg=f"{url} {desc}")



import unittest

from rethink.sso.github import GithubSSO


class SSOTest(unittest.IsolatedAsyncioTestCase):
    async def test_github(self):
        sso = GithubSSO(
            client_id="",
            client_secret="",
            redirect_uri="",
            allow_insecure_http=False,
            use_state=False,
            scope=None,
        )
        self.assertEqual("github", sso.provider)
        self.assertEqual({
            "authorization_endpoint": "https://github.com/login/oauth/authorize",
            "token_endpoint": "https://github.com/login/oauth/access_token",
            "userinfo_endpoint": "https://api.github.com/user",
        }, await sso.get_discovery_document())

    async def test_github_openid_from_response(self):
        sso = GithubSSO(
            client_id="",
            client_secret="",
            redirect_uri="",
            allow_insecure_http=False,
            use_state=False,
            scope=None,
        )
        open_id = await sso.openid_from_response({
            "email": "email",
            "id": "id",
            "login": "login",
            "avatar_url": "avatar_url",
        })
        self.assertEqual("id", open_id.id)
        self.assertEqual("github", open_id.provider)
        self.assertEqual("login", open_id.display_name)
        self.assertEqual("avatar_url", open_id.picture)
        self.assertEqual("email", open_id.email)



import re
from pathlib import Path
from typing import List

import jieba.analyse
import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity

STOP_WORDS_FILE = Path(__file__).parent / "data" / "stop_words.txt"
IDF_FILE = Path(__file__).parent / "data" / "idf.txt"
ENG_IDF_FILE = Path(__file__).parent / "data" / "wiki_idf.txt"


class TFIDF:
    drop_pattern = re.compile(r"[^\u4e00-\u9fa5a-zA-Z0-9 ]")

    def __init__(self):
        self.stop_words, idf, vocabulary = self._load_dict()
        self.vectorizer = TfidfVectorizer(
            tokenizer=self.tokenizer,
            token_pattern=None,
            dtype=np.float32,
        )
        self.vectorizer.idf_ = idf
        self.vectorizer.vocabulary_ = vocabulary

    @staticmethod
    def _load_dict():
        sw = set(STOP_WORDS_FILE.read_text(encoding="utf-8").splitlines())
        lines = IDF_FILE.read_text(encoding="utf-8").splitlines()
        lines.extend(ENG_IDF_FILE.read_text(encoding="utf-8").splitlines())
        vocabulary = {}
        idf = np.empty(len(lines), dtype=np.float32)
        for i, line in enumerate(lines):
            w, f = line.split(" ")
            vocabulary[w] = i
            idf[i] = float(f)
        return sw, idf, vocabulary

    def tokenizer(self, text: str) -> List[str]:
        text = text.lower()
        text = self.drop_pattern.sub("", text)
        for w in jieba.cut_for_search(text):
            if w.strip() == "":
                continue
            if w in self.stop_words:
                continue
            yield w

    def docs2vec(self, docs: List[str]) -> np.ndarray:
        return self.vectorizer.transform(docs)

    def search(self, query: str, docs_vec: np.ndarray, n: int = 5) -> List[int]:
        """
        search top n similar docs

        Args:
            query (str): search text
            docs_vec (np.ndarray): candidate docs in vector
            n (int): return top n similar docs

        Returns:
            List[int]: top n similar docs index
        """
        q_vec = self.vectorizer.transform([query])
        cs = cosine_similarity(docs_vec, q_vec)
        top_n = cs.ravel().argsort()[-n:][::-1]
        return top_n

# tfidf = TFIDF()



import datetime
import io
import os
import threading
import zipfile
from typing import List, Tuple, Optional

import pymongo.errors
import requests
from bson import ObjectId
from bson.tz_util import utc
from fastapi import UploadFile
from starlette.datastructures import Headers

from rethink import const, models
from rethink.config import is_local_db
from rethink.logger import logger
from rethink.models.database import COLL
from rethink.models.tps import ImportData
from . import file_ops

MAX_IMAGE_SIZE = 1024 * 1024 * 10  # 10 mb
RESIZE_IMG_THRESHOLD = 1024 * 128  # 128 kb


def update_process(
        uid: str,
        type_: str,
        process: int,
        start_at: datetime.datetime = None,
        running: bool = None,
        code: int = None,
) -> Tuple[Optional[ImportData], const.Code]:
    data = {"type": type_, "process": process}
    if start_at is not None:
        data["startAt"] = start_at
    if running is not None:
        data["running"] = running
    if code is not None:
        data["code"] = code
    if is_local_db():
        # local db not support find_one_and_update
        COLL.import_data.update_one({"uid": uid}, {"$set": data})
        doc = COLL.import_data.find_one({"uid": uid})
    else:
        doc = COLL.import_data.find_one_and_update(
            {"uid": uid},
            {"$set": data}
        )
    if doc is None:
        return doc, const.Code.OPERATION_FAILED
    return doc, const.Code.OK


def __set_running_false(
        uid: str,
        code: const.Code,
        problem_files: List[str] = None,
) -> None:
    COLL.import_data.update_one({"uid": uid}, {"$set": {
        "running": False,
        "problemFiles": problem_files if problem_files is not None else [],
        "code": code.value,
    }})


def upload_obsidian_thread(
        uid: str,
        bytes_data: bytes,
        filename: str,
        doc: dict,
        max_file_size: int,
) -> None:
    try:
        unzipped_files = file_ops.unzip_file(bytes_data)
    except zipfile.BadZipFile:
        __set_running_false(uid, const.Code.INVALID_FILE_TYPE, [filename])
        logger.info(f"invalid file type: {filename}, uid: {uid}")
        return
    filtered_files = {}
    existed_filename2nid = doc["obsidian"].copy()
    img_path_dict = {}
    img_name_dict = {}
    md_count = 0
    for filepath, data in unzipped_files.items():
        try:
            base_name, ext = filepath.rsplit(".", 1)
        except ValueError:
            continue
        valid_file = {"md", "txt"}
        valid_file.update(file_ops.VALID_IMG_EXT)
        if ext not in valid_file:
            continue
        if data["size"] > max_file_size:
            __set_running_false(uid, const.Code.TOO_LARGE_FILE, [filepath], )
            logger.info(f"too large file: {filepath}, uid: {uid}")
            return
        if ext in ["md", "txt"]:
            if len(filepath.split("/")) > 1:
                continue
            md_count += 1
            filtered_files[filepath] = data["file"]
        else:
            img_path_dict[filepath] = data["file"]
            img_name_dict[os.path.basename(filepath)] = data["file"]

    # add new md files
    for i, (filepath, file_bytes) in enumerate(filtered_files.items()):
        base_name, ext = filepath.rsplit(".", 1)
        if base_name in existed_filename2nid:
            continue
        try:
            n, code = models.node.add(
                uid=uid,
                md=base_name,
                type_=const.NodeType.MARKDOWN.value,
            )
        except pymongo.errors.DuplicateKeyError:
            continue
        if code != const.Code.OK:
            __set_running_false(uid, code, [filepath], )
            logger.info(f"error: {code}, filepath: {filepath}, uid: {uid}")
            return
        existed_filename2nid[base_name] = n["id"]
        if i % 20 == 0:
            doc, code = update_process(uid, "obsidian", int(i / md_count * 10))
            if code != const.Code.OK:
                __set_running_false(uid, code, [filepath], )
                logger.info(f"error: {code}, filepath: {filepath}, uid: {uid}")
                return
            if not doc["running"]:
                break

    # update all files
    for i, (filepath, file_bytes) in enumerate(filtered_files.items()):
        base_name, ext = filepath.rsplit(".", 1)
        try:
            md = file_bytes.decode("utf-8")
        except (FileNotFoundError, OSError, UnicodeDecodeError) as e:
            logger.error(f"error: {e}. filepath: {filepath}")
            __set_running_false(uid, const.Code.FILE_OPEN_ERROR, [filepath], )
            logger.info(f"error: {const.Code.FILE_OPEN_ERROR}, filepath: {filepath}, uid: {uid}")
            return

        md = file_ops.replace_inner_link_and_upload_image(
            uid=uid,
            md=md,
            exist_filename2nid=existed_filename2nid,
            img_path_dict=img_path_dict,
            img_name_dict=img_name_dict,
            resize_threshold=RESIZE_IMG_THRESHOLD,
        )
        md = base_name + "\n\n" + md
        nid = existed_filename2nid[base_name]
        n, code = models.node.update(
            uid=uid,
            nid=nid,
            md=md,
            refresh_on_same_md=True,
        )
        if code == const.Code.NODE_NOT_EXIST:
            n, code = models.node.add(
                uid=uid,
                md=md,
                type_=const.NodeType.MARKDOWN.value,
            )
            if code != const.Code.OK:
                __set_running_false(uid, code, [filepath])
                logger.info(f"error: {code}, filepath: {filepath}, uid: {uid}")
                return
        elif code != const.Code.OK:
            __set_running_false(uid, code, [filepath])
            logger.info(f"error: {code}, filepath: {filepath}, uid: {uid}")
            return
        if i % 20 == 0:
            doc, code = update_process(uid, "obsidian", int(i / md_count * 40 + 10))
            if code != const.Code.OK:
                __set_running_false(uid, code, [filepath])
                logger.info(f"error: {code}, filepath: {filepath}, uid: {uid}")
                return
            if not doc["running"]:
                break

    # update for old obsidian files
    count = 0
    for base_name, nid in doc["obsidian"].items():
        if base_name not in existed_filename2nid:
            n, code = models.node.get(uid=uid, nid=nid)
            if code != const.Code.OK:
                continue
            n, code = models.node.update(
                uid=uid,
                nid=nid,
                md=n["md"],
                refresh_on_same_md=True,
            )
            if code != const.Code.OK:
                continue
        if count % 20 == 0:
            doc, code = update_process(uid, "obsidian", int(count / len(doc["obsidian"]) * 50 + 50))
            if code != const.Code.OK:
                __set_running_false(uid, code)
                logger.info(f"error: {code}, uid: {uid}")
                return
            if not doc["running"]:
                break

        count += 1

    COLL.import_data.update_one(
        {"uid": uid},
        {"$set": {
            "obsidian": existed_filename2nid,
            "running": False,
            "problemFiles": [],
            "code": 0,
            "process": 100,
        }})


def upload_obsidian(uid: str, zipped_files: List[UploadFile]) -> const.Code:
    max_file_count = 1
    max_file_size = 1024 * 1024 * 200  # 200 mb

    doc = COLL.import_data.find_one({"uid": uid})
    if doc is None:
        doc: ImportData = {
            "_id": ObjectId(),
            "uid": uid,
            "process": 0,
            "type": "obsidian",
            "startAt": datetime.datetime.now(tz=utc),
            "running": True,
            "problemFiles": [],
            "code": 0,
            "obsidian": {},
        }
        res = COLL.import_data.insert_one(doc)
        if not res.acknowledged:
            __set_running_false(uid, const.Code.OPERATION_FAILED)
            return const.Code.OPERATION_FAILED
    elif doc["running"]:
        return const.Code.IMPORT_PROCESS_NOT_FINISHED
    _, code = update_process(
        uid=uid,
        type_="obsidian",
        process=0,
        start_at=datetime.datetime.now(tz=utc),
        running=True,
        code=0,
    )
    if code != const.Code.OK:
        __set_running_false(uid, code)
        return code

    # check file type and size and number
    if len(zipped_files) > max_file_count:
        __set_running_false(uid, const.Code.TOO_MANY_FILES)
        return const.Code.TOO_MANY_FILES

    zipped_file = zipped_files[0]
    if not zipped_file.filename.endswith(".zip"):
        __set_running_false(uid, const.Code.INVALID_FILE_TYPE)
        return const.Code.INVALID_FILE_TYPE
    if zipped_file.content_type not in ["application/zip", "application/octet-stream", "application/x-zip-compressed"]:
        __set_running_false(uid, const.Code.INVALID_FILE_TYPE)
        return const.Code.INVALID_FILE_TYPE

    td = threading.Thread(
        target=upload_obsidian_thread,
        args=(uid, zipped_file.file.read(), zipped_file.filename, doc, max_file_size),
        daemon=True,
    )
    td.start()
    return const.Code.OK


def update_text_thread(
        uid: str,
        files: List[dict],
        max_file_size: int,
):
    for file in files:
        if not file["filename"].endswith(".md") and not file["filename"].endswith(".txt"):
            __set_running_false(uid, const.Code.INVALID_FILE_TYPE, [file["filename"]])
            return
        if file["size"] > max_file_size:
            __set_running_false(uid, const.Code.TOO_LARGE_FILE, [file["filename"]])
            return

    for i, file in enumerate(files):
        try:
            md = file["content"].decode("utf-8")
        except (FileNotFoundError, OSError) as e:
            logger.error(f"error: {e}. filepath: {file['filename']}")
            __set_running_false(uid, const.Code.FILE_OPEN_ERROR, [file["filename"]])
            return
        title = file["filename"].rsplit(".", 1)[0]
        md = title + "\n\n" + md
        try:
            n, code = models.node.add(
                uid=uid,
                md=md,
                type_=const.NodeType.MARKDOWN.value,
            )
        except pymongo.errors.DuplicateKeyError:
            continue
        if code != const.Code.OK:
            __set_running_false(uid, code, [file["filename"]])
            return
        if i % 20 == 0:
            doc, code = update_process(uid, "text", int(i / len(files) * 100))
            if code != const.Code.OK:
                __set_running_false(uid, code, [file["filename"]])
                return
            if not doc["running"]:
                break
    COLL.import_data.update_one(
        {"uid": uid},
        {"$set": {
            "running": False,
            "process": 100,
            "code": 0,
            "problemFiles": [],
        }})


def upload_text(uid: str, files: List[UploadFile]) -> const.Code:
    max_file_count = 200
    max_file_size = 1024 * 512  # 512 kb
    doc = COLL.import_data.find_one({"uid": uid})
    if doc is None:
        doc: ImportData = {
            "_id": ObjectId(),
            "uid": uid,
            "process": 0,
            "type": "text",
            "startAt": datetime.datetime.now(tz=utc),
            "running": True,
            "problemFiles": [],
            "code": 0,
            "obsidian": {},
        }
        res = COLL.import_data.insert_one(doc)
        if not res.acknowledged:
            __set_running_false(uid, const.Code.OPERATION_FAILED)
            return const.Code.OPERATION_FAILED
    elif doc["running"]:
        return const.Code.IMPORT_PROCESS_NOT_FINISHED

    _, code = update_process(
        uid,
        "text",
        0,
        start_at=datetime.datetime.now(tz=utc),
        running=True,
        code=0,
    )
    if code != const.Code.OK:
        __set_running_false(uid, code)
        return code

    if len(files) > max_file_count:
        __set_running_false(uid, const.Code.TOO_MANY_FILES)
        return const.Code.TOO_MANY_FILES

    file_list = [{
        "filename": file.filename,
        "content": file.file.read(),
        "size": file.size,
    } for file in files]

    td = threading.Thread(
        target=update_text_thread,
        args=(uid, file_list, max_file_size),
        daemon=True,
    )
    td.start()
    return const.Code.OK


def get_upload_process(uid: str) -> Optional[dict]:
    timeout_minus = 5
    doc = COLL.import_data.find_one({"uid": uid})
    if doc is None:
        return None
    now = datetime.datetime.now(tz=utc)

    # upload timeout
    if doc["running"] and \
            now.replace(tzinfo=None) - doc["startAt"].replace(tzinfo=None) \
            > datetime.timedelta(minutes=timeout_minus):
        doc["running"] = False
        COLL.import_data.update_one(
            {"uid": uid},
            {"$set": {
                "running": False,
                "code": const.Code.UPLOAD_TASK_TIMEOUT.value,
                "problemFiles": [],
            }},
        )
    return doc


def upload_image_vditor(uid: str, files: List[UploadFile]) -> dict:
    res = {
        "errFiles": [],
        "succMap": {},
        "code": const.Code.OK,
    }
    u, code = models.user.get(uid=uid)
    if code != const.Code.OK:
        res["errFiles"] = [file.filename for file in files]
        res["code"] = code
        return res
    if models.user.user_space_not_enough(u=u):
        res["errFiles"] = [file.filename for file in files]
        res["code"] = const.Code.USER_SPACE_NOT_ENOUGH
        return res

    return file_ops.save_upload_files(
        uid=uid,
        files=files,
        max_image_size=MAX_IMAGE_SIZE,
        resize_threshold=RESIZE_IMG_THRESHOLD,
    )


def fetch_image_vditor(uid: str, url: str) -> Tuple[str, const.Code]:
    u, code = models.user.get(uid=uid)
    if code != const.Code.OK:
        return "", code
    if models.user.user_space_not_enough(u=u):
        return "", const.Code.USER_SPACE_NOT_ENOUGH

    try:
        r = requests.get(url)
    except requests.exceptions.RequestException:
        return url, const.Code.OK

    if r.status_code != 200:
        return "", const.Code.FILE_OPEN_ERROR
    file = UploadFile(
        filename=url.split("/")[-1],
        file=io.BytesIO(r.content),
        headers=Headers(r.headers),
        size=len(r.content)
    )
    res = file_ops.save_upload_files(
        uid=uid,
        files=[file],
        max_image_size=MAX_IMAGE_SIZE,
        resize_threshold=RESIZE_IMG_THRESHOLD,
    )
    if len(res["errFiles"]) > 0:
        return "", const.Code.FILE_OPEN_ERROR
    return res["succMap"][file.filename], const.Code.OK



import datetime
import shutil
import unittest
from io import BytesIO
from pathlib import Path
from textwrap import dedent

import requests
from PIL import Image
from bson import ObjectId
from bson.tz_util import utc
from fastapi import UploadFile
from starlette.datastructures import Headers

from rethink import const, models
from rethink.models.utils import short_uuid
from . import utils


class LocalModelsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")
        models.database.drop_all()

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")

    def setUp(self) -> None:
        models.database.init()
        u, _ = models.user.get_by_email(email=const.DEFAULT_USER["email"])
        self.uid = u["id"]

    def tearDown(self) -> None:
        models.database.drop_all()
        shutil.rmtree(Path(__file__).parent / "tmp" / ".data" / "files", ignore_errors=True)
        shutil.rmtree(Path(__file__).parent / "tmp" / ".data" / "md", ignore_errors=True)

    def test_user(self):
        _id, code = models.user.add(
            account="aaa", source=const.UserSource.EMAIL.value,
            email="aaa", hashed="bbb", nickname="ccc", avatar="ddd", language=const.Language.EN.value)
        self.assertNotEqual("", _id)
        self.assertEqual(const.Code.OK, code)

        u, code = models.user.get_by_email(email=const.DEFAULT_USER["email"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("rethink", u["nickname"])
        self.assertIsNotNone(u)

        u, code = models.user.get(_id)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("ccc", u["nickname"])

        u, code = models.user.update(
            uid=_id,
            email="a@a.com",
            hashed="1",
            nickname="2",
            avatar="3",
            language="zh",
            node_display_method=const.NodeDisplayMethod.LIST.value,
        )
        self.assertEqual(const.Code.OK, code)

        u, code = models.user.get(_id)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("a@a.com", u["email"])
        self.assertEqual("1", u["hashed"])
        self.assertEqual("2", u["nickname"])
        self.assertEqual("3", u["avatar"])
        self.assertEqual("zh", u["language"])
        self.assertEqual(const.NodeDisplayMethod.LIST.value, u["lastState"]["nodeDisplayMethod"])

        code = models.user.disable(uid=_id)
        self.assertEqual(const.Code.OK, code)

        u, code = models.user.get(uid=_id)
        self.assertEqual(const.Code.ACCOUNT_OR_PASSWORD_ERROR, code)
        self.assertIsNone(u)

        code = models.user.enable(uid=_id)
        self.assertEqual(const.Code.OK, code)

        code = models.user.disable(uid="sdwqdqw")
        self.assertEqual(const.Code.OPERATION_FAILED, code)

        code = models.user.delete(uid=_id)
        self.assertEqual(const.Code.OK, code)

    def test_node(self):
        node, code = models.node.add(
            uid=self.uid, md="a" * (const.MD_MAX_LENGTH + 1), type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.NOTE_EXCEED_MAX_LENGTH, code)
        self.assertIsNone(node)

        u, code = models.user.get(self.uid)
        self.assertEqual(const.Code.OK, code)
        used_space = u["usedSpace"]
        node, code = models.node.add(
            uid=self.uid, md="[title](/qqq)\nbody", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        u, code = models.user.get(self.uid)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(used_space + len(node["md"].encode("utf-8")), u["usedSpace"])
        self.assertEqual("modifiedAt", u["lastState"]["nodeDisplaySortKey"])

        n, code = models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("title", n["title"])
        self.assertEqual("body", n["snippet"])

        ns, total = models.search.user_node(uid=self.uid)
        self.assertEqual(3, len(ns))
        self.assertEqual(3, total)

        ns, total = models.search.user_node(uid=self.uid, page_size=5, page=12, sort_key="createdAt")
        self.assertEqual(0, len(ns))
        self.assertEqual(3, total)

        u, code = models.user.update(
            uid=self.uid,
            node_display_sort_key="createdAt",
        )

        self.assertEqual(const.Code.OK, code)
        self.assertEqual("createdAt", u["lastState"]["nodeDisplaySortKey"])
        used_space = u["usedSpace"]
        n, code = models.node.update(uid=self.uid, nid=node["id"], md="title2\nbody2")
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("title2", n["title"])
        self.assertEqual("title2\nbody2", n["md"])
        self.assertEqual(const.NodeType.MARKDOWN.value, n["type"])

        u, code = models.user.get(self.uid)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(used_space + (
                len(n["md"].encode("utf-8")) -
                len(node["md"].encode("utf-8"))
        ), u["usedSpace"])

        code = models.node.disable(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        n, code = models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.NODE_NOT_EXIST, code)

        code = models.node.to_trash(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)

        code = models.node.delete(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        n, code = models.node.get(uid=self.uid, nid=node["id"])
        self.assertIsNone(n)
        self.assertEqual(const.Code.NODE_NOT_EXIST, code)

        u, code = models.user.get(self.uid)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(used_space - len(node["md"].encode("utf-8")), u["usedSpace"])

    def test_parse_at(self):
        nid1, _ = models.node.add(
            uid=self.uid, md="c", type_=const.NodeType.MARKDOWN.value,
        )
        nid2, _ = models.node.add(
            uid=self.uid, md="我133", type_=const.NodeType.MARKDOWN.value,
        )
        md = dedent(f"""title
        fffqw [@c](/n/{nid1['id']})
        fff 
        [@我133](/n/{nid2['id']})
        ffq
        """)
        node, code = models.node.add(
            uid=self.uid, md=md, type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        nodes, total = models.search.user_node(uid=self.uid)
        self.assertEqual(5, len(nodes))
        self.assertEqual(5, total)
        found, total = models.search.user_node(uid=self.uid, query="我")
        self.assertEqual(2, len(found), msg=found)
        self.assertEqual(5, total)

        n, code = models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(2, len(n["toNodeIds"]))

        cache = n["md"]
        n, code = models.node.update(uid=self.uid, nid=node["id"], md=f'{cache}xxxx')
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(cache + "xxxx", n["md"])

        n, code = models.node.get(uid=self.uid, nid=nid1['id'])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(1, len(n["fromNodeIds"]))

        n, code = models.node.update(uid=self.uid, nid=node["id"], md=n["title"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(0, len(n["toNodeIds"]))

        n, code = models.node.get(uid=self.uid, nid=nid1['id'])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(0, len(n["fromNodeIds"]))

    def test_add_set(self):
        node, code = models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(0, len(node["toNodeIds"]))
        self.assertEqual(const.Code.OK, code)

        res = models.db_ops.node_add_to_set(node["id"], "toNodeIds", short_uuid())
        self.assertEqual(1, res.modified_count)
        node, code = models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(1, len(node["toNodeIds"]))

    def test_cursor_text(self):
        n1, code = models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        n2, code = models.node.add(
            uid=self.uid, md="title2\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)

        recom = models.node.cursor_query(
            uid=self.uid,
            nid=n2["id"],
            cursor_text="te",
        )
        self.assertEqual(2, len(recom))

        recom = models.node.cursor_query(
            uid=self.uid,
            nid=n2["id"],  # exclude the second node
            cursor_text="",
        )
        self.assertEqual(2, len(recom))

        code = models.search.add_recent_cursor_search(self.uid, n1["id"], n2["id"])
        self.assertEqual(const.Code.OK, code)

        recom = models.node.cursor_query(
            uid=self.uid,
            nid=n1["id"],  # exclude the second node
            cursor_text="",
        )
        self.assertEqual(3, len(recom))
        self.assertEqual("Welcome to Rethink", recom[2]["title"])

    def test_to_trash(self):
        n1, code = models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        n2, code = models.node.add(
            uid=self.uid, md="title2\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)

        code = models.node.to_trash(self.uid, n1["id"])
        self.assertEqual(const.Code.OK, code)

        ns, total = models.node.get_nodes_in_trash(self.uid, 0, 10)
        self.assertEqual(1, len(ns))
        self.assertEqual(1, total)
        self.assertEqual(n1["id"], ns[0]["id"])

        ns, total = models.search.user_node(self.uid)
        self.assertEqual(3, len(ns))
        self.assertEqual(3, total)

        code = models.node.restore_from_trash(self.uid, n1["id"])
        self.assertEqual(const.Code.OK, code)
        nodes, total = models.search.user_node(self.uid)
        self.assertEqual(4, len(nodes))
        self.assertEqual(4, total)

    def test_search(self):
        code = models.search.put_recent_search(self.uid, "a")
        self.assertEqual(const.Code.OK, code)
        models.search.put_recent_search(self.uid, "c")
        models.search.put_recent_search(self.uid, "b")

        queries = models.search.get_recent_search(self.uid)
        self.assertEqual(["b", "c", "a"], queries)

    def test_batch(self):
        ns = []
        for i in range(10):
            n, code = models.node.add(
                uid=self.uid, md=f"title{i}\ntext", type_=const.NodeType.MARKDOWN.value
            )
            self.assertEqual(const.Code.OK, code)
            ns.append(n)

        base_count = 2

        code = models.node.batch_to_trash(self.uid, [n["id"] for n in ns[:4]])
        self.assertEqual(const.Code.OK, code)
        nodes, total = models.search.user_node(self.uid)
        self.assertEqual(6 + base_count, len(nodes))
        self.assertEqual(6 + base_count, total)

        tns, total = models.node.get_nodes_in_trash(self.uid, 0, 10)
        self.assertEqual(4, total)
        self.assertEqual(4, len(tns))

        code = models.node.restore_batch_from_trash(self.uid, [n["id"] for n in tns[:2]])
        self.assertEqual(const.Code.OK, code)
        nodes, total = models.search.user_node(self.uid)
        self.assertEqual(8 + base_count, len(nodes))
        self.assertEqual(8 + base_count, total)

        code = models.node.batch_delete(self.uid, [n["id"] for n in tns[2:4]])
        self.assertEqual(const.Code.OK, code)
        tns, total = models.node.get_nodes_in_trash(self.uid, 0, 10)
        self.assertEqual(0, total)
        self.assertEqual(0, len(tns))

    def test_files_upload_process(self):
        now = datetime.datetime.now(tz=utc)
        doc: models.tps.ImportData = {
            "_id": ObjectId(),
            "uid": "xxx",
            "process": 0,
            "type": "text",
            "startAt": now,
            "running": True,
            "obsidian": {},
            "problemFiles": [],
            "code": const.Code.OK.value,
        }
        res = models.database.COLL.import_data.insert_one(doc)
        self.assertTrue(res.acknowledged)

        doc, code = models.files.update_process("xxx", "obsidian", 10)
        self.assertEqual(const.Code.OK, code)

        doc = models.files.get_upload_process("xxx")
        self.assertEqual(10, doc["process"])
        self.assertEqual("obsidian", doc["type"])
        self.assertEqual(now, doc["startAt"])
        self.assertTrue(doc["running"])

        models.database.COLL.import_data.delete_one({"uid": "xxx"})

    def test_update_title_and_from_nodes_updates(self):
        n1, code = models.node.add(
            uid=self.uid, md=f"title1\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        n2, code = models.node.add(
            uid=self.uid, md=f"title2\n[@title1](/n/{n1['id']})", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)

        n1, code = models.node.update(uid=self.uid, nid=n1["id"], md="title1Changed\ntext")
        self.assertEqual(const.Code.OK, code)
        n2, code = models.node.get(uid=self.uid, nid=n2["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(f"title2\n[@title1Changed](/n/{n1['id']})", n2["md"])

    def test_upload_image_vditor(self):
        u, code = models.user.get(self.uid)
        used_space = u["usedSpace"]

        p = Path(__file__).parent.parent / "img" / "phone-notes.png"
        image = Image.open(p)
        buf = BytesIO()
        image.save(buf, format="png")
        size = buf.tell()
        img_file = UploadFile(
            buf, filename="phone-notes.png", size=size,
            headers=Headers({"content-type": "image/png"})
        )
        res = models.files.upload_image_vditor(self.uid, [img_file])
        self.assertIn("phone-notes.png", res["succMap"])
        self.assertTrue(".png" in res["succMap"]["phone-notes.png"])
        local_file = Path(__file__).parent / "tmp" / ".data" / res["succMap"]["phone-notes.png"][1:]
        self.assertTrue(local_file.exists())
        local_file.unlink()

        u, code = models.user.get(self.uid)
        self.assertEqual(used_space + size, u["usedSpace"])

    def test_fetch_image_vditor(self):
        u, code = models.user.get(self.uid)
        used_space = u["usedSpace"]

        url = "https://rethink.run/favicon.ico"
        new_url, code = models.files.fetch_image_vditor(self.uid, url)
        self.assertEqual(const.Code.OK, code)
        self.assertTrue(new_url.endswith(".ico"))
        self.assertTrue(new_url.startswith("/"))
        local_file = Path(__file__).parent / "tmp" / ".data" / new_url[1:]
        self.assertTrue(local_file.exists())
        local_file.unlink()

        u, code = models.user.get(self.uid)
        r = requests.get(url)
        self.assertEqual(used_space + len(r.content), u["usedSpace"])

    def test_update_used_space(self):
        u, code = models.user.get(self.uid)
        base_used_space = u["usedSpace"]
        for delta, value in [
            (100, 100),
            (100, 200),
            (0, 200),
            (-300, 0),
            (20.1, 20.1),
        ]:
            code = models.user.update_used_space(self.uid, delta)
            self.assertEqual(const.Code.OK, code)
            u, code = models.user.get(self.uid)
            self.assertEqual(const.Code.OK, code)
            self.assertAlmostEqual(value, u["usedSpace"] - base_used_space, msg=f"delta: {delta}, value: {value}")



import datetime
import io
import os
import shutil
import time
import unittest
from pathlib import Path
from typing import Dict
from zipfile import ZipFile

from PIL import Image
from fastapi.testclient import TestClient
from httpx import Response

from rethink import const
from rethink.application import app
from rethink.models import database
from rethink.models.utils import jwt_decode
from . import utils


class PublicApiTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")
        database.init()

    def setUp(self) -> None:
        self.client = TestClient(app)

    @classmethod
    def tearDownClass(cls) -> None:
        database.drop_all()
        utils.drop_env(".env.test.local")

    def test_home(self):
        resp = self.client.get("/")
        self.assertEqual(200, resp.status_code)
        self.assertEqual("text/html; charset=utf-8", resp.headers["content-type"])
        self.assertEqual(Response, type(resp))

    def test_register(self):
        resp = self.client.put("/api/user", json={
            "email": "a@c.com",
            "password": "a",
            "language": const.Language.EN.value,
            "requestId": "xxx"
        })
        rj = resp.json()
        self.assertEqual(const.Code.ONE_USER_MODE.value, rj["code"])
        self.assertEqual(len(rj["token"]), 0)
        self.assertEqual("xxx", rj["requestId"])


class TokenApiTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")

    def setUp(self) -> None:
        database.init()
        self.client = TestClient(app)
        resp = self.client.post("/api/login", json={
            "email": const.DEFAULT_USER["email"],
            "password": "",
        })
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.token = rj["token"]

    def tearDown(self) -> None:
        database.drop_all()
        shutil.rmtree(Path(__file__).parent / "tmp" / ".data" / "files", ignore_errors=True)
        shutil.rmtree(Path(__file__).parent / "tmp" / ".data" / "md", ignore_errors=True)

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")

    def test_get_user(self):
        resp = self.client.get(
            "/api/user",
            headers={"token": self.token, "rid": "xxx"})
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("rethink", rj["user"]["nickname"])
        self.assertEqual("xxx", rj["requestId"])
        self.assertGreater(
            datetime.datetime.strptime(rj["user"]["createdAt"], "%Y-%m-%dT%H:%M:%SZ").timestamp(),
            0., msg=rj["user"]["createdAt"])

    def test_update_user(self):
        resp = self.client.post("/api/user", json={
            "requestId": "xxx",
            "nickname": "new nickname",
            "avatar": "http://new.avatar/aa.png",
            "nodeDisplayMethod": const.NodeDisplayMethod.LIST.value,
        }, headers={"token": self.token})
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])

        resp = self.client.get(
            "/api/user",
            headers={"token": self.token, "rid": "xxx"}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("new nickname", rj["user"]["nickname"])
        self.assertEqual("http://new.avatar/aa.png", rj["user"]["avatar"])
        self.assertEqual(const.NodeDisplayMethod.LIST.value, rj["user"]["lastState"]["nodeDisplayMethod"])
        self.assertEqual("xxx", rj["requestId"])

    def test_recent_search(self):
        resp = self.client.put(
            "/api/search/recent",
            json={
                "requestId": "xxx",
                "query": "aaa",
            }, headers={"token": self.token})
        rj = resp.json()
        self.assertEqual(0, rj["code"])

        resp = self.client.get(
            "/api/search/recent",
            headers={"token": self.token, "rid": "xxx"}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])
        self.assertEqual(["aaa"], rj["queries"])

    def test_node(self):
        resp = self.client.post(
            "/api/search/node",
            json={
                "query": "",
                "requestId": "xxx",
                "sortKey": "createdAt",
                "sortOrder": -1, "page": 0, "pageSize": 5},
            headers={"token": self.token})
        rj = resp.json()
        self.assertGreater(len(rj["data"]["nodes"]), 0)

        self.client.post(
            "/api/search/node",
            json={
                "query": "",
                "requestId": "xxx",
                "sortKey": "createdAt",
                "sortOrder": -1, "page": 0, "pageSize": 5},
            headers={"token": self.token})

        resp = self.client.put("/api/node", json={
            "requestId": "xxx",
            "md": "node1\ntext",
            "type": const.NodeType.MARKDOWN.value,
        }, headers={"token": self.token})
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])
        node = rj["node"]

        resp = self.client.get(
            "/api/node",
            params={"nid": node["id"]},
            headers={"token": self.token, "rid": "xxx"}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])
        n = rj["node"]
        self.assertEqual("node1", n["title"], msg=rj)
        self.assertEqual("node1\ntext", n["md"])
        self.assertEqual(const.NodeType.MARKDOWN.value, n["type"])

        resp = self.client.post(
            "/api/node",
            json={
                "requestId": "xxx",
                "nid": node["id"],
                "md": "node2\ntext"
            },
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])

        resp = self.client.get(
            "/api/node",
            params={"nid": node["id"]},
            headers={"token": self.token, "rid": "xxx"}
        )
        rj = resp.json()
        n = rj["node"]
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])
        self.assertEqual("node2", n["title"])
        self.assertEqual("node2\ntext", n["md"])
        self.assertEqual(const.NodeType.MARKDOWN.value, n["type"])

        resp = self.client.put(
            "/api/trash",
            json={
                "requestId": "xxx",
                "nid": node["id"],
            },
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])

        resp = self.client.get(
            "/api/trash",
            params={"p": 0, "ps": 10},
            headers={"token": self.token, "rid": "xxx"}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual(1, len(rj["data"]["nodes"]))

        resp = self.client.post(
            "/api/trashRestore",
            json={
                "requestId": "xxx",
                "nid": node["id"],
            },
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])

        resp = self.client.post(
            "/api/search/cursor",
            json={
                "requestId": "xxx",
                "nid": node["id"],
                "textBeforeCursor": "How",
            },
            headers={"token": self.token},
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])
        self.assertEqual(1, len(rj["nodes"]))
        self.assertEqual("How do I record", rj["nodes"][0]["title"])

        self.client.put(
            "/api/trash",
            json={
                "requestId": "xxx",
                "nid": node["id"],
            },
            headers={"token": self.token}
        )
        resp = self.client.delete(
            f"/api/trash/{node['id']}",
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])

        resp = self.client.delete(
            "/api/trash/ssa",
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(const.Code.NODE_NOT_EXIST.value, rj["code"])

        resp = self.client.get(
            "/api/node",
            params={"nid": node["id"]},
            headers={"token": self.token, "rid": "xxx"}
        )
        rj = resp.json()
        self.assertEqual(const.Code.NODE_NOT_EXIST.value, rj["code"])

    def test_captcha(self):
        resp = self.client.get(
            "/api/captcha/img",
        )
        self.assertEqual(200, resp.status_code)
        self.assertEqual("image/png", resp.headers["content-type"])
        self.assertGreater(len(resp.headers["x-captcha-token"]), 0)

        code_str = jwt_decode(resp.headers["x-captcha-token"])["code"]
        resp = self.client.post(
            "/api/captcha/img",
            json={
                "requestId": "xxx",
                "token": resp.headers["x-captcha-token"],
                "codeStr": code_str,
            }
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])

    def test_batch(self):
        base_count = 2
        for i in range(10):
            resp = self.client.put("/api/node", json={
                "requestId": "xxx",
                "md": f"node{i}\ntext",
                "type": const.NodeType.MARKDOWN.value,
            }, headers={"token": self.token})
            rj = resp.json()
            self.assertEqual(0, rj["code"])
        resp = self.client.post(
            "/api/search/node",
            json={
                "query": "",
                "requestId": "xxx",
                "sortKey": "createdAt",
                "sortOrder": -1, "page": 0, "pageSize": 5},
            headers={"token": self.token})
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual(5, len(rj["data"]["nodes"]))
        self.assertEqual(10 + base_count, rj["data"]["total"])

        resp = self.client.put(
            "/api/trash/batch",
            json={
                "requestId": "xxx",
                "nids": [n["id"] for n in rj["data"]["nodes"][:3]],
            },
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        resp = self.client.get(
            "/api/trash",
            params={"p": 0, "ps": 10},
            headers={"token": self.token, "rid": "xxx"}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual(3, len(rj["data"]["nodes"]))
        self.assertEqual(3, rj["data"]["total"])

        resp = self.client.post(
            "/api/trashRestore/batch",
            json={
                "requestId": "xxx",
                "nids": [n["id"] for n in rj["data"]["nodes"][:2]],
            },
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        resp = self.client.get(
            "/api/trash",
            params={"p": 0, "ps": 10},
            headers={"token": self.token, "rid": "xxx"}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual(1, len(rj["data"]["nodes"]))
        self.assertEqual(1, rj["data"]["total"])

        resp = self.client.post(
            "/api/trashDelete/batch",
            json={
                "requestId": "xxx",
                "nids": [n["id"] for n in rj["data"]["nodes"]],
            },
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        resp = self.client.get(
            "/api/trash",
            params={"p": 0, "ps": 10},
            headers={"token": self.token, "rid": "xxx"}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual(0, len(rj["data"]["nodes"]))
        self.assertEqual(0, rj["data"]["total"])

    def test_update_obsidian(self):
        image = Image.new('RGB', (4, 4))
        img_byte_arr = io.BytesIO()
        image.save(img_byte_arr, format='PNG')
        orig_data: Dict[str, bytes] = {
            "test1.md": "111\n[[qa]] ![img/p.png](img/p.png)".encode("utf-8"),
            "test2.md": "111\n\nasdq[[test]] ![[Pasted image 20230810112909.png]]".encode("utf-8"),
            "test3.md": "111\n\n![[Pasted image 20230810112931.png]]".encode("utf-8"),
            "20230810112909.png": img_byte_arr.getvalue(),
            os.path.join("img", "p.png"): img_byte_arr.getvalue(),
            os.path.join("img", "20230810112931.png"): img_byte_arr.getvalue(),
        }
        # write dummy zip file
        zip_bytes = ZipFile("test.zip", "w")
        for k, v in orig_data.items():
            zip_bytes.writestr(k, v)
        zip_bytes.close()
        f = open("test.zip", "rb")

        resp = self.client.post(
            "/api/files/obsidian",
            files=[
                ("files", f),
            ],
            headers={
                "token": self.token
            },
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"], msg=rj)

        for _ in range(10):
            time.sleep(0.1)
            resp = self.client.get(
                "/api/files/uploadProcess",
                headers={"token": self.token, "rid": "xxx"}
            )
            rj = resp.json()
            self.assertEqual("obsidian", rj["type"], msg=rj)
            self.assertEqual([], rj["problemFiles"], msg=rj)
            if rj["process"] == 100:
                break
        self.assertFalse(rj["running"], msg=rj)

        resp = self.client.post(
            "/api/search/node",
            json={
                "query": "",
                "requestId": "xxx",
                "sortKey": "createdAt",
                "sortOrder": -1, "page": 0, "pageSize": 5},
            headers={"token": self.token})
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual(5, len(rj["data"]["nodes"]))
        self.assertEqual(5, rj["data"]["total"])

        f.close()
        os.remove("test.zip")

    def test_upload_text(self):
        os.makedirs("temp", exist_ok=True)
        f1 = open("temp/test1.txt", "wb+")
        f1.write("dasd".encode("utf-8"))
        f2 = open("temp/test2.txt", "wb+")
        f2.write("asdq".encode("utf-8"))

        resp = self.client.post(
            "/api/files/text",
            files=[
                ("files", f1),
                ("files", f2),
            ],
            headers={
                "token": self.token
            },
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])

        for _ in range(8):
            time.sleep(0.1)
            resp = self.client.get(
                "/api/files/uploadProcess",
                headers={"token": self.token, "rid": "xxx"}
            )
            rj = resp.json()
            self.assertEqual("text", rj["type"], msg=rj)
            self.assertEqual([], rj["problemFiles"], msg=rj)
            if rj["process"] == 100:
                break
        self.assertFalse(rj["running"], msg=rj)

        resp = self.client.post(
            "/api/search/node",
            json={
                "query": "",
                "requestId": "xxx",
                "sortKey": "createdAt",
                "sortOrder": -1, "page": 0, "pageSize": 5},
            headers={"token": self.token})
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual(4, len(rj["data"]["nodes"]))

        f1.close()
        f2.close()
        os.remove("temp/test1.txt")
        os.remove("temp/test2.txt")
        os.rmdir("temp")

    def test_upload_image(self):
        os.makedirs("temp", exist_ok=True)
        image = Image.new('RGB', (4, 4))
        image.save("temp/test.png", format='PNG')
        f1 = open("temp/test.png", "rb")
        resp = self.client.post(
            "/api/files/imageUploadVditor",
            files={"file[]": f1},
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual({
            'errFiles': [],
            'succMap': {
                'test.png': f"/files/3acca26d4f9d111694d7dbda2d1e6a40.png"
            }}, rj["data"])
        f1.close()
        os.remove("temp/test.png")
        os.rmdir("temp")

    def test_put_quick_node(self):
        resp = self.client.put(
            "/api/node/quick",
            json={
                "requestId": "xxx",
                "md": "node1\ntext",
                "type": const.NodeType.MARKDOWN.value,
            },
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])
        node = rj["node"]
        self.assertEqual("node1", node["title"])
        self.assertEqual("node1\ntext", node["md"])
        self.assertEqual(const.NodeType.MARKDOWN.value, node["type"])

        resp = self.client.put(
            "/api/node/quick",
            json={
                "requestId": "xxx",
                "md": "https://baidu.com",
                "type": const.NodeType.MARKDOWN.value,
            },
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])
        node = rj["node"]
        self.assertIn("https://baidu.com", node["md"])
        self.assertIn("百度", node["md"])



import os

from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import HTMLResponse, FileResponse
from fastapi.staticfiles import StaticFiles

from rethink import const, config
from rethink.logger import logger
from .models import database
from .routes import (
    user,
    oauth,
    node,
    search,
    trash,
    verification,
    files,
)

app = FastAPI(
    docs_url="/docs",
    openapi_url="/openapi.json",
)

if os.getenv("VUE_APP_MODE", "local") not in ["development", "local"]:
    allow_origins = [
        "https://rethink.run",
        "https://www.rethink.run",
    ]
else:
    allow_origins = [
        "*",
    ]

app.add_middleware(
    CORSMiddleware,
    allow_origins=allow_origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
    expose_headers=["X-Captcha-Token"],
)

app.include_router(node.router)
app.include_router(user.router)
app.include_router(oauth.router)
app.include_router(search.router)
app.include_router(trash.router)
app.include_router(verification.router)
app.include_router(files.router)


@app.on_event("startup")
async def startup_event():
    database.init()
    logger.info("db initialized")


app.mount(
    "/css",
    StaticFiles(directory=const.FRONTEND_DIR / "css"),
    name="css",
)
app.mount(
    "/js",
    StaticFiles(directory=const.FRONTEND_DIR / "js"),
    name="js",
)
app.mount(
    "/img",
    StaticFiles(directory=const.FRONTEND_DIR / "img"),
    name="img",
)


@app.get("/", response_class=HTMLResponse)
@app.get("/login", response_class=HTMLResponse)
@app.get("/about", response_class=HTMLResponse)
@app.get("/r", response_class=HTMLResponse)
@app.get("/r/{path}", response_class=HTMLResponse)
@app.get("/n/{nid}", response_class=HTMLResponse)
async def index() -> HTMLResponse:
    content = (const.FRONTEND_DIR / "index.html").read_text(encoding="utf-8")
    content += f"<script>window.VUE_APP_API_PORT = {os.getenv('VUE_APP_API_PORT')}</script>"
    return HTMLResponse(
        content=content,
        status_code=200,
    )


@app.get("/files/{fid}", response_class=FileResponse)
async def user_data(
        fid: str,
) -> FileResponse:
    if config.is_local_db():
        prefix = config.get_settings().LOCAL_STORAGE_PATH
    else:
        raise HTTPException(status_code=404, detail="only support local storage")
    return FileResponse(
        path=prefix / ".data" / "files" / fid,
        status_code=200,
    )



import calendar
import datetime
import json
import os
import struct
from dataclasses import dataclass
from typing import Optional, Union

from bson import ObjectId
from bson.tz_util import utc
from mongita import MongitaClientDisk
from mongita.collection import Collection
from pymongo import MongoClient
from pymongo.collection import Collection as RemoteCollection

from rethink import config, const
from rethink.logger import logger
from . import utils
from .tps import UserMeta, Node, UserFile, ImportData


@dataclass
class Collections:
    users: Union[Collection, RemoteCollection] = None
    nodes: Union[Collection, RemoteCollection] = None
    import_data: Union[Collection, RemoteCollection] = None
    user_file: Union[Collection, RemoteCollection] = None


COLL = Collections()
CLIENT: Optional[Union[MongoClient, MongitaClientDisk]] = None


def set_client():
    global CLIENT
    conf = config.get_settings()
    if config.is_local_db():
        if not conf.LOCAL_STORAGE_PATH.exists():
            raise FileNotFoundError(f"Path not exists: {conf.LOCAL_STORAGE_PATH}")
        db_path = conf.LOCAL_STORAGE_PATH / ".data" / "db"
        db_path.mkdir(parents=True, exist_ok=True)
        CLIENT = MongitaClientDisk(db_path)
    else:
        CLIENT = MongoClient(
            host=conf.DB_HOST,
            port=conf.DB_PORT,
            username=conf.DB_USER,
            password=conf.DB_PASSWORD,
        )


def __remote_try_build_index():
    # try creating index
    if not isinstance(CLIENT, MongoClient):
        return

    users_info = COLL.users.index_information()
    if "id_1" not in users_info:
        COLL.users.create_index("id", unique=True)
    if "account_1_source_1" not in users_info:
        COLL.users.create_index(["account", "source"], unique=True)

    nodes_info = COLL.nodes.index_information()
    if "id_1" not in nodes_info:
        COLL.nodes.create_index("id", unique=True)
    if "uid_1_id_-1_modifiedAt_-1" not in nodes_info:
        COLL.nodes.create_index(
            [("uid", 1), ("id", -1), ("modifiedAt", -1)],
            unique=True,
        )
    if "uid_1_id_-1_inTrash_-1" not in nodes_info:
        COLL.nodes.create_index(
            [("uid", 1), ("id", -1), ("inTrash", -1)],
            unique=True,
        )
    if "uid_1_id_-1" not in nodes_info:
        # created at
        COLL.nodes.create_index(
            [("uid", 1), ("id", -1)],
            unique=True,
        )
    if "uid_1_id_-1_title_1" not in nodes_info:
        COLL.nodes.create_index(
            [("uid", 1), ("id", -1), ("title", 1)],
            unique=True,
        )

    import_data_info = COLL.import_data.index_information()
    if "uid_1" not in import_data_info:
        COLL.import_data.create_index("uid", unique=True)

    user_file_info = COLL.user_file.index_information()
    if "uid_1_fid_-1" not in user_file_info:
        COLL.user_file.create_index([("uid", 1), ("fid", -1)], unique=True)


def init():
    set_client()
    db = CLIENT[config.get_settings().DB_NAME]
    COLL.users = db["users"]
    COLL.nodes = db["nodes"]
    COLL.import_data = db["importData"]
    COLL.user_file = db["userFile"]

    if config.is_local_db():
        __local_try_create_or_restore()
    else:
        __remote_try_build_index()


def get_client():
    return CLIENT


def drop_all():
    set_client()
    for db_name in CLIENT.list_database_names():
        if db_name == "admin":
            continue
        CLIENT.drop_database(db_name)


def __local_try_add_default_user():
    dot_rethink_path = config.get_settings().LOCAL_STORAGE_PATH / ".data" / ".rethink.json"
    u_insertion = {
        "_id": ObjectId(),
        "id": utils.short_uuid(),
        "email": const.DEFAULT_USER["email"],
        "nickname": const.DEFAULT_USER["nickname"],
        "avatar": const.DEFAULT_USER["avatar"],
        "account": const.DEFAULT_USER["email"],
        "language": os.getenv("VUE_APP_LANGUAGE", const.Language.EN.value),
    }
    with open(dot_rethink_path, "w", encoding="utf-8") as f:
        out = u_insertion.copy()
        out["_id"] = str(out["_id"])
        json.dump(out, f, indent=2, ensure_ascii=False)

    logger.info("running at the first time, a user with initial data will be created")
    ns = const.NEW_USER_DEFAULT_NODES[u_insertion["language"]]

    def create_node(md: str):
        title_, snippet_ = utils.preprocess_md(md)
        n: Node = {
            "_id": ObjectId(),
            "id": utils.short_uuid(),
            "uid": u_insertion["id"],
            "title": title_,
            "snippet": snippet_,
            "md": md,
            "type": const.NodeType.MARKDOWN.value,
            "disabled": False,
            "inTrash": False,
            "modifiedAt": datetime.datetime.now(tz=utc),
            "inTrashAt": None,
            "fromNodeIds": [],
            "toNodeIds": [],
            "searchKeys": utils.txt2search_keys(title_),
        }
        res = COLL.nodes.insert_one(n)
        if not res.acknowledged:
            raise ValueError("cannot insert default node")
        return n

    n0 = create_node(ns[0])
    _md = ns[1].format(n0["id"])
    n1 = create_node(_md)

    u: UserMeta = {
        "_id": ObjectId(u_insertion["_id"]),
        "id": u_insertion["id"],
        "email": u_insertion["email"],
        "nickname": u_insertion["nickname"],
        "avatar": u_insertion["avatar"],
        "account": u_insertion["email"],
        "language": u_insertion["language"],

        "source": const.UserSource.LOCAL.value,
        "hashed": "",
        "disabled": False,
        "modifiedAt": datetime.datetime.now(tz=utc),
        "usedSpace": 0,
        "type": const.USER_TYPE.NORMAL.id,
        "lastState": {
            "recentCursorSearchSelectedNIds": [n0["id"], n1["id"]],
            "recentSearch": [],
            "nodeDisplayMethod": const.NodeDisplayMethod.CARD.value,
            "nodeDisplaySortKey": "modifiedAt"
        }
    }
    _ = COLL.users.insert_one(u)


def __oid_from_datetime(dt: datetime.datetime) -> ObjectId:
    offset = dt.utcoffset()
    if offset is not None:
        dt = dt - offset
    timestamp = calendar.timegm(dt.timetuple())
    oid = struct.pack(">I", timestamp)
    oid += ObjectId._random()
    with ObjectId._inc_lock:
        oid += struct.pack(">I", ObjectId._inc)[1:4]
        ObjectId._inc = (ObjectId._inc + 1) % (0xFFFFFF + 1)
    return ObjectId(oid)


def __local_restore():
    # restore user
    dot_rethink_path = config.get_settings().LOCAL_STORAGE_PATH / ".data" / ".rethink.json"
    if not dot_rethink_path.exists():
        return
    with open(dot_rethink_path, "r", encoding="utf-8") as f:
        u_insertion = json.load(f)

    u: UserMeta = {
        "_id": ObjectId(u_insertion["_id"]),
        "id": u_insertion["id"],
        "email": u_insertion["email"],
        "nickname": u_insertion["nickname"],
        "avatar": u_insertion["avatar"],
        "account": u_insertion["email"],
        "language": u_insertion["language"],

        "source": const.UserSource.LOCAL.value,
        "hashed": "",
        "disabled": False,
        "modifiedAt": datetime.datetime.now(tz=utc),
        "usedSpace": 0,
        "type": const.USER_TYPE.NORMAL.id,
        "lastState": {
            "recentCursorSearchSelectedNIds": [],
            "recentSearch": [],
            "nodeDisplayMethod": const.NodeDisplayMethod.CARD.value,
            "nodeDisplaySortKey": "modifiedAt"
        }
    }
    _ = COLL.users.insert_one(u)

    # restore nodes
    md_dir = config.get_settings().LOCAL_STORAGE_PATH / ".data" / "md"
    if not md_dir.exists():
        return
    ns = []
    for md_path in md_dir.glob("*.md"):
        md = md_path.read_text(encoding="utf-8")
        created_time = datetime.datetime.fromtimestamp(md_path.stat().st_ctime, tz=utc)
        modified_time = datetime.datetime.fromtimestamp(md_path.stat().st_mtime, tz=utc)
        title_, snippet_ = utils.preprocess_md(md)

        n: Node = {
            "_id": __oid_from_datetime(created_time),
            "id": md_path.stem,
            "uid": u_insertion["id"],
            "title": title_,
            "snippet": snippet_,
            "md": md,
            "type": const.NodeType.MARKDOWN.value,
            "disabled": False,
            "inTrash": False,
            "modifiedAt": modified_time,
            "inTrashAt": None,
            "fromNodeIds": [],
            "toNodeIds": [],
            "searchKeys": utils.txt2search_keys(title_),
        }
        ns.append(n)
    res = COLL.nodes.insert_many(ns)
    if not res.acknowledged:
        raise ValueError("cannot insert default node")
    logger.info(f"restore nodes count: {len(res.inserted_ids)}")

    docs = []
    for f in config.get_settings().LOCAL_STORAGE_PATH.glob(".data/files/*"):
        filename = f.name
        fid = filename.split(".")[0]
        created_time = datetime.datetime.fromtimestamp(f.stat().st_ctime, tz=utc)
        docs.append({
            "_id": __oid_from_datetime(created_time),
            "uid": u_insertion["id"],
            "fid": fid,
            "filename": filename,
            "size": f.stat().st_size,
        })
    res = COLL.user_file.insert_many(docs)
    if not res.acknowledged:
        raise ValueError("cannot insert default node")
    logger.info(f"restore files count: {len(res.inserted_ids)}")


def __local_try_create_or_restore():
    if not config.get_settings().ONE_USER:
        return

    # check if field changes
    for c, t in [
        (COLL.users, UserMeta),
        (COLL.nodes, Node),
        (COLL.user_file, UserFile),
        (COLL.import_data, ImportData),
    ]:
        doc = c.find_one()
        if doc:
            if set(doc.keys()) != set(t.__annotations__):
                drop_all()
                __local_restore()
                return
    # check local files count matches db
    md_dir = config.get_settings().LOCAL_STORAGE_PATH / ".data" / "md"
    if md_dir.exists():
        if len(list(md_dir.glob("*.md"))) != COLL.nodes.count_documents({}):
            drop_all()
            __local_restore()
            return
    files_dir = config.get_settings().LOCAL_STORAGE_PATH / ".data" / "files"
    if files_dir.exists():
        if len(list(files_dir.glob("*"))) != COLL.user_file.count_documents({}):
            drop_all()
            __local_restore()
            return

    # try fix TypeError: can't compare offset-naive and offset-aware datetimes
    docs = COLL.nodes.find()
    for doc in docs:
        COLL.nodes.update_one(
            {"_id": doc["_id"]},
            {"$set": {"modifiedAt": doc["modifiedAt"].replace(tzinfo=utc)}},
        )
    docs = COLL.import_data.find()
    for doc in docs:
        COLL.import_data.update_one(
            {"_id": doc["_id"]},
            {"$set": {"startAt": doc["startAt"].replace(tzinfo=utc)}},
        )

    if COLL.users.find_one(
            {"account": const.DEFAULT_USER["email"], "source": const.UserSource.EMAIL.value}
    ) is not None:
        return

    # no default user, create one
    __local_try_add_default_user()



from typing import List, Sequence

from pydantic import Field, BaseModel, NonNegativeInt

from .node import NodesInfoResponse


class SearchUserNodesRequest(BaseModel):
    requestId: str
    query: str = ""
    sortKey: str = "createdAt"
    sortOrder: int = -1
    page: NonNegativeInt = 0
    pageSize: NonNegativeInt = 0
    nidExclude: Sequence[str] = Field(default_factory=list)


class CursorQueryRequest(BaseModel):
    nid: str
    textBeforeCursor: str
    requestId: str = ""


class CursorQueryResponse(BaseModel):
    code: NonNegativeInt
    message: str
    requestId: str
    nodes: List[NodesInfoResponse.Data.NodeInfo]


class AddToRecentSearchHistRequest(BaseModel):
    requestId: str
    nid: str
    toNid: str


class PutRecentSearchRequest(BaseModel):
    requestId: str
    query: str


class GetRecentSearchResponse(BaseModel):
    code: NonNegativeInt
    message: str
    requestId: str
    queries: List[str]



import unittest

import bcrypt

from rethink import const
from rethink.controllers import auth
from rethink.models import database
from . import utils


class AuthTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        utils.set_env(".env.test.local")
        database.init()
        cls.salt = bcrypt.gensalt(rounds=5)

    @classmethod
    def tearDownClass(cls) -> None:
        database.drop_all()
        utils.drop_env(".env.test.local")

    def test_verify(self):
        password = "123abc"
        email = "rethink@rethink.run"
        bpw = auth._base_password(password=password, email=email)
        token_str = bcrypt.hashpw(bpw, self.salt).decode("utf-8")
        hashed = auth._base_password(password=password, email=email)
        match = bcrypt.checkpw(hashed, token_str.encode("utf-8"))
        self.assertTrue(match)

    def test_hash(self):
        pwd = "123abc&&rethink@rethink.run"
        pwd_bt = pwd.encode("utf-8")
        hashed = bcrypt.hashpw(pwd_bt, self.salt)
        # print(hashed)
        match = bcrypt.checkpw(pwd_bt, hashed)
        self.assertTrue(match)

    def test_one_user(self):
        uid, code = auth.register_user("a@q.com", "rethink", const.Language.EN.value)
        self.assertEqual(const.Code.ONE_USER_MODE, code)
        self.assertEqual("", uid)

    def test_verify_user(self):
        u, err = auth.get_user_by_email("rethink@rethink.run")
        self.assertEqual(const.Code.OK, err)
        ok = auth.verify_user(u, "rethink")
        self.assertTrue(ok)

    def test_valid_password(self):
        for t, b in [
            ("rethink", False),
            ("123334", False),
            ("ret123", True),
            ("re123", False),
            ("rethink123", True),
            ("sd@d", False),
            ("sd@d123", True),
            ("1s", False),
            ("", False)
        ]:
            if b:
                self.assertIsNotNone(auth.VALID_PASSWORD_PTN.match(t), msg=t)
            else:
                self.assertIsNone(auth.VALID_PASSWORD_PTN.match(t), msg=t)



from datetime import timedelta
from io import BytesIO
from random import randint
from typing import Tuple, Dict

import jwt
from captcha.audio import AudioCaptcha
from captcha.image import ImageCaptcha

from rethink import const
from .utils import alphabet, alphabet_len, jwt_encode, jwt_decode

DEFAULT_CAPTCHA_EXPIRE_SECOND = 60

img_captcha = ImageCaptcha(font_sizes=(35, 30, 32))
audio_captcha = AudioCaptcha()


def random_captcha(length: int = 4, sound: bool = False) -> Tuple[str, Dict[str, BytesIO]]:
    code = [alphabet[randint(0, alphabet_len - 1)] for _ in range(length)]
    code_str = "".join(code)
    data = {
        "img": img_captcha.generate(code_str),
    }
    if sound:
        data["sound"] = audio_captcha.generate(code_str)
    token = jwt_encode(
        exp_delta=timedelta(seconds=DEFAULT_CAPTCHA_EXPIRE_SECOND),
        data={"code": code_str.lower()}
    )
    return token, data


def verify_captcha(token: str, code_str: str) -> const.Code:
    code = const.Code.CAPTCHA_ERROR
    try:
        data = jwt_decode(token)
        if data["code"] == code_str.lower():
            code = const.Code.OK
    except jwt.ExpiredSignatureError:
        code = const.Code.CAPTCHA_EXPIRED
    except (jwt.DecodeError, Exception):
        code = const.Code.INVALID_AUTH
    return code



from typing import Any, Dict, List

from mongita.results import UpdateResult

from rethink import config
from .database import COLL
from .tps import Node


def remove_from_node(from_nid: str, to_nid: str):
    if config.is_local_db():
        # no $pull support
        to_n = COLL.nodes.find_one({"id": to_nid})
        if to_n is None:
            return
        try:
            to_n["fromNodeIds"].remove(from_nid)
            COLL.nodes.update_one(
                {"id": to_nid},
                {"$set": {"fromNodeIds": to_n["fromNodeIds"]}}
            )
        except ValueError:
            pass
    else:
        COLL.nodes.update_one(
            {"id": to_nid},
            {"$pull": {"fromNodeIds": from_nid}}
        )


def node_add_to_set(id_: str, key: str, value: Any) -> UpdateResult:
    res = UpdateResult(0, 0)
    if config.is_local_db():
        # no $addToSet support
        has_new = False
        doc = COLL.nodes.find_one({"id": id_})
        if doc is None:
            return res
        if key not in doc:
            doc[key] = []
        if value not in doc[key]:
            doc[key].append(value)
            has_new = True
        if has_new:
            res = COLL.nodes.update_one(
                {"id": id_},
                {"$set": {key: doc[key]}}
            )
    else:
        res = COLL.nodes.update_one(
            {"id": id_},
            {"$addToSet": {key: value}}
        )
    return res


def nodes_get(
        uid: str,
        ids: List[str],
        assert_conditions: Dict[str, Any]
) -> List[Node]:
    c = {"id": {"$in": ids}, "uid": uid, **assert_conditions}
    docs = list(COLL.nodes.find(c))
    return docs



import datetime
import io
import math
import re
import uuid
from typing import Tuple

import httpx
import jwt
import pypinyin
from markdown import Markdown

from rethink import config
from rethink.logger import logger

HEADERS = {
    'typ': 'jwt',
    'alg': 'RS256'
}
alphabet = "23456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
alphabet_len = len(alphabet)
__padding = int(math.ceil(math.log(2 ** 128, alphabet_len)))
__code_pattern = re.compile(r"^```[^\S\r\n]*[a-z]*?\n(.*?)\n```$", re.MULTILINE | re.DOTALL)


def short_uuid() -> str:
    """
    The output has the most significant digit first.
    """
    number = uuid.uuid4().int
    output = ""

    while number:
        number, digit = divmod(number, alphabet_len)
        output += alphabet[digit]
    if __padding:
        remainder = max(__padding - len(output), 0)
        output = output + alphabet[0] * remainder
    return output[::-1]


def __unmark_element(element, stream=None):
    if stream is None:
        stream = io.StringIO()
    if element.text:
        stream.write(element.text)
    for sub in element:
        __unmark_element(sub, stream)
    if element.tail:
        stream.write(element.tail)
    return stream.getvalue()


# patching Markdown
Markdown.output_formats["plain"] = __unmark_element
__md = Markdown(output_format="plain")
__md.stripTopLevelTags = False


def md2txt(md: str) -> str:
    for found in list(__code_pattern.finditer(md))[::-1]:
        span = found.span()
        code = found.group(1)
        md = f"{md[: span[0]]}{code}{md[span[1]:]}"
    return __md.convert(md)


def preprocess_md(md: str) -> Tuple[str, str]:
    title, body = split_title_body(fulltext=md)
    title = md2txt(title.strip())
    body = body.strip()
    snippet = md2txt(md=body)[:200]
    return title, snippet


def txt2search_keys(txt: str) -> str:
    s1 = "".join(pypinyin.lazy_pinyin(txt))
    s2 = "".join(pypinyin.lazy_pinyin(txt, style=pypinyin.Style.BOPOMOFO))
    s = {s1, s2, txt}
    return " ".join(s).lower()


def split_title_body(fulltext: str) -> (str, str):
    title_body = fulltext.split("\n", maxsplit=1)
    title = title_body[0].strip()
    try:
        body = title_body[1].strip()
    except IndexError:
        body = ""
    return title, body


def jwt_encode(exp_delta: datetime.timedelta, data: dict) -> str:
    """
    Create token

    Args:
        exp_delta: expired delta
        data: payload data

    Returns:
        str: token
    """
    payload = {
        "exp": datetime.datetime.utcnow() + exp_delta
    }
    payload.update(data)
    token = jwt.encode(
        payload=payload,
        key=config.get_settings().JWT_KEY,
        algorithm=HEADERS["alg"],
        headers=HEADERS,
    )
    return token


def jwt_decode(token: str) -> dict:
    return jwt.decode(
        token,
        key=config.get_settings().JWT_KEY_PUB,
        algorithms=[HEADERS["alg"]],
        options={"verify_exp": True}
    )


def change_link_title(md: str, nid: str, new_title: str) -> str:
    new_md = re.sub(
        r"\[@[^].]*?]\(/n/{}/?\)".format(nid),
        f"[@{new_title}](/n/{nid})",
        md,
    )
    return new_md


ONLY_HTTP_LINK_PTN = re.compile(r"^https?://\S*$")


def contain_only_http_link(md: str) -> str:
    content = md.strip()
    if ONLY_HTTP_LINK_PTN.match(content) is None:
        return ""
    return content


ASYNC_CLIENT_HEADERS = {
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,"
              "application/signed-exchange;v=b3;q=0.7",
    "Accept-Encoding": "gzip, deflate, br",
    "Accept-Language": "zh",
    "Cache-Control": "no-cache",
    "Pragma": "no-cache",
    "Sec-Ch-Ua": '"Chromium";v="118", "Google Chrome";v="118", "Not=A?Brand";v="99"',
    "Sec-Ch-Ua-Mobile": "?0",
    "Sec-Ch-Ua-Platform": '"macOS"',
    "Sec-Fetch-Dest": "document",
    "Sec-Fetch-Mode": "navigate",
    "Sec-Fetch-Site": "none",
    "Sec-Fetch-User": "?1",
    "Upgrade-Insecure-Requests": "1",
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
                  "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
}


async def get_title_description_from_link(url: str) -> Tuple[str, str]:
    async with httpx.AsyncClient() as client:
        try:
            response = await client.get(
                url=url,
                headers=ASYNC_CLIENT_HEADERS,
                timeout=3.
            )
        except (
                httpx.ConnectTimeout,
                RuntimeError,
                httpx.ConnectError,
                httpx.ReadTimeout,
                httpx.HTTPError
        ) as e:
            logger.info(f"failed to get {url}: {e}")
            return "", ""
        if response.status_code in [302, 301]:
            url = response.headers["Location"]
            return await get_title_description_from_link(url)
        if response.status_code != 200:
            return "", ""
        html = response.text

    title, description = "", ""
    found = re.search(r'<meta[^>]*name="title"[^>]*content="([^"]*)"[^>]*>', html, re.DOTALL)
    if found is None:
        found = re.search(r'<meta[^>]*content="([^"]*)"[^>]*name="title"[^>]*>', html, re.DOTALL)
    if found is None:
        found = re.search(r"<title[^>]*>(.*?)</title>", html, re.DOTALL)
    if found:
        title = found.group(1).strip()

    found = re.search(r'<meta[^>]*name="description"[^>]*content="([^"]*)"[^>]*>', html, re.DOTALL)
    if found is None:
        found = re.search(r'<meta[^>]*content="([^"]*)"[^>]*name="description"[^>]*>', html, re.DOTALL)
    if found:
        description = found.group(1).strip()[:400]
    return title, description



from rethink import const, models
from rethink.controllers import schemas
from rethink.controllers.utils import TokenDecode, datetime2str


def cursor_query(
        td: TokenDecode,
        req: schemas.search.CursorQueryRequest,
) -> schemas.search.CursorQueryResponse:
    if td.code != const.Code.OK:
        return schemas.search.CursorQueryResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId=req.requestId,
            nodes=None
        )
    recommended_nodes = models.node.cursor_query(
        uid=td.uid,
        nid=req.nid,
        cursor_text=req.textBeforeCursor,
    )
    code = const.Code.OK
    return schemas.search.CursorQueryResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId=req.requestId,
        nodes=[schemas.node.NodesInfoResponse.Data.NodeInfo(
            id=n["id"],
            title=n["title"],
            snippet=n["snippet"],
            type=n["type"],
            createdAt=datetime2str(n["_id"].generation_time),
            modifiedAt=datetime2str(n["modifiedAt"]),
        ) for n in recommended_nodes],
    )


def search_user_nodes(
        td: TokenDecode,
        req: schemas.search.SearchUserNodesRequest,
) -> schemas.node.NodesInfoResponse:
    if td.code != const.Code.OK:
        return schemas.node.NodesInfoResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId=req.requestId,
            nodes=[],
        )
    nodes, total = models.search.user_node(
        uid=td.uid,
        query=req.query,
        sort_key=req.sortKey,
        sort_order=req.sortOrder,
        page=req.page,
        page_size=req.pageSize,
        nid_exclude=req.nidExclude,
    )
    code = const.Code.OK
    return schemas.node.NodesInfoResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId=req.requestId,
        data=schemas.node.parse_nodes_info(nodes, total),
    )


def add_to_recent_cursor_search(
        td: TokenDecode,
        req: schemas.search.AddToRecentSearchHistRequest
) -> schemas.base.AcknowledgeResponse:
    if td.code != const.Code.OK:
        return schemas.base.AcknowledgeResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId=req.requestId,
        )
    code = models.search.add_recent_cursor_search(uid=td.uid, nid=req.nid, to_nid=req.toNid)
    return schemas.base.AcknowledgeResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId=req.requestId,
    )


def get_recent(
        td: TokenDecode,
        rid: str,
) -> schemas.search.GetRecentSearchResponse:
    if td.code != const.Code.OK:
        return schemas.search.GetRecentSearchResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId=rid,
            queries=[],
        )
    queries = models.search.get_recent_search(uid=td.uid)
    code = const.Code.OK
    return schemas.search.GetRecentSearchResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId=rid,
        queries=queries,
    )


def put_recent(
        td: TokenDecode,
        req: schemas.search.PutRecentSearchRequest,
) -> schemas.base.AcknowledgeResponse:
    if td.code != const.Code.OK:
        return schemas.base.AcknowledgeResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId=req.requestId,
        )
    code = models.search.put_recent_search(uid=td.uid, query=req.query)
    return schemas.base.AcknowledgeResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId=req.requestId,
    )



import hashlib
import io
import re
import zipfile
from pathlib import Path
from platform import system
from typing import List, Tuple, BinaryIO, Dict, Union

from PIL import Image, UnidentifiedImageError
from bson.objectid import ObjectId
from fastapi import UploadFile
from qcloud_cos import CosConfig, CosServiceError, CosS3Client

from rethink import const, config
from rethink.models.database import COLL
from rethink.models.tps import UserFile
from rethink.models.user import update_used_space
from rethink.models.utils import short_uuid

INTERNAL_LINK_PTN = re.compile(r"\[\[(.*?)]]")
INTERNAL_IMG_PTN = re.compile(r"!\[\[(Pasted image .*?)]]")
INTERNAL_IMG_PTN2 = re.compile(r"!\[(.*?)]\((?!http)(.*?)\)")

VALID_IMG_EXT = {"jpg", "jpeg", "png", "gif", "svg"}


def delete_file(uid: str, fid: str) -> const.Code:
    success_deletion, failed_deletion = delete_files(uid=uid, fids=[fid])
    if len(success_deletion) != 1:
        return const.Code.OPERATION_FAILED
    return const.Code.OK


def delete_files(uid: str, fids: List[str]) -> Tuple[List[str], List[str]]:
    success_deletion = []
    failed_deletion = []
    delta = 0
    for fid in fids:
        doc = COLL.user_file.find_one_and_delete({"uid": uid, "fid": fid})
        if doc is None:
            failed_deletion.append(fid)
        else:
            success_deletion.append(fid)
            delta -= doc["size"]

    update_used_space(uid=uid, delta=delta)
    return success_deletion, failed_deletion


def unzip_file(zip_bytes: bytes) -> Dict[str, Dict[str, Union[bytes, int]]]:
    with zipfile.ZipFile(io.BytesIO(zip_bytes), 'r') as ref:
        filepaths = ref.namelist()
        extracted_files = {}
        for filepath in filepaths:
            info = ref.getinfo(filepath)
            try:
                if system() in ["Darwin", "Linux"]:
                    _filepath = filepath.encode('cp437').decode('utf-8')
                elif system() == "Windows":
                    _filepath = filepath.encode('cp437').decode('gbk')
                else:
                    _filepath = filepath
            except UnicodeEncodeError:
                if system() == "Windows":
                    _filepath = filepath.encode("utf-8").decode("utf-8")
                _filepath = filepath
            sp = _filepath.split("/")
            if sp[0] in ["__MACOSX", ".DS_Store"]:
                continue
            if len(sp) > 1:
                _filepath = "/".join(sp[1:])
            if _filepath.strip() == "" or _filepath.startswith("."):
                continue
            extracted_files[_filepath] = {
                "file": ref.read(filepath),
                "size": info.file_size,
            }

    return extracted_files


def __img_ptn_replace_upload(
        uid: str,
        filepath: str,
        filename: str,
        md: str,
        img_dict: Dict[str, bytes],
        resize_threshold: int,
        span: Tuple[int, int]
) -> str:
    """

    Args:
        filepath:
        filename:
        md:
        img_dict:
        resize_threshold:
        span:

    Returns:
        md, uploaded_file_size
    """
    if filepath not in img_dict:
        return md
    file = img_dict[filepath]
    ext = filepath.split(".")[-1].lower()
    if ext == "jpg":
        ext = "jpeg"
    content_type = f"image/{ext}"
    bio_file = io.BytesIO(file)
    bio_file.seek(0, io.SEEK_END)
    file_size_ = bio_file.tell()
    url = __save_image(
        uid=uid,
        filename=filepath,
        file=bio_file,
        file_size=file_size_,
        content_type=content_type,
        resize_threshold=resize_threshold,
    )
    if url != "":
        md = f"{md[: span[0]]}![{filename}]({url}){md[span[1]:]}"
    return md


def replace_inner_link_and_upload_image(
        uid: str,
        md: str,
        exist_filename2nid: Dict[str, str],
        img_path_dict: Dict[str, bytes],
        img_name_dict: Dict[str, bytes],
        resize_threshold: int,
) -> str:
    """

    Args:
        uid:
        md:
        exist_filename2nid:
        img_path_dict:
        img_name_dict:
        resize_threshold:

    Returns:
        md
    """
    # image
    for match in list(INTERNAL_IMG_PTN2.finditer(md))[::-1]:
        span = match.span()
        filename = match.group(1)
        filepath = match.group(2)
        md = __img_ptn_replace_upload(
            uid=uid,
            filepath=filepath,
            filename=filename,
            md=md,
            img_dict=img_path_dict,
            resize_threshold=resize_threshold,
            span=span,
        )
    for match in list(INTERNAL_IMG_PTN.finditer(md))[::-1]:
        span = match.span()
        filepath = filename = match.group(1)
        md = __img_ptn_replace_upload(
            uid=uid,
            filepath=filepath,
            filename=filename,
            md=md,
            img_dict=img_name_dict,
            resize_threshold=resize_threshold,
            span=span,
        )

    # link
    for match in list(INTERNAL_LINK_PTN.finditer(md))[::-1]:
        span = match.span()
        filename = match.group(1)
        nid = exist_filename2nid.get(filename)
        if nid is None:
            nid = short_uuid()
            exist_filename2nid[filename] = nid
        md = f"{md[: span[0]]}[@{filename}](/n/{nid}){md[span[1]:]}"
    return md


def file_hash(file: BinaryIO) -> str:
    md5_hash = hashlib.md5()
    file.seek(0)
    for chunk in iter(lambda: file.read(4096), b""):
        md5_hash.update(chunk)
    file.seek(0)
    return md5_hash.hexdigest()


def __get_out_bytes(
        file_size: int,
        content_type: str,
        file: BinaryIO,
        resize_threshold: int,
) -> BinaryIO:
    if file_size > resize_threshold:
        # reduce image size
        out = io.BytesIO()
        image = Image.open(file)
        image.save(out, format=content_type.split("/")[-1].upper(), quality=50, optimize=True)
    else:
        out = file
    out.seek(0)
    return out


def __save_image(
        uid: str,
        filename: str,
        file: BinaryIO,
        file_size: int,
        content_type: str,
        resize_threshold: int,
) -> str:
    """

    Args:
        filename:
        file:
        file_size:
        content_type:
        resize_threshold:

    Returns:
        url
    """
    fn = Path(filename)
    ext = fn.suffix.lower()
    hashed = file_hash(file)

    if config.is_local_db():
        url = upload_to_local_storage(
            uid=uid,
            file_size=file_size,
            content_type=content_type,
            file=file,
            resize_threshold=resize_threshold,
            filename=filename,
            fid=hashed + ext,
        )

    else:
        # upload to cos
        try:
            url = upload_bytes_to_cos(
                uid=uid,
                file_size=file_size,
                content_type=content_type,
                file=file,
                resize_threshold=resize_threshold,
                filename=filename,
                fid=hashed + ext,
            )
        except CosServiceError:
            return ""

    return url


def save_upload_files(
        uid: str,
        files: List[UploadFile],
        max_image_size: int,
        resize_threshold: int,
) -> dict:
    res = {
        "errFiles": [],
        "succMap": {},
        "code": const.Code.OK,
    }
    for file in files:
        filename = file.filename
        if not file.content_type.startswith("image/"):
            res["errFiles"].append(filename)
            continue
        if file.size > max_image_size:
            res["errFiles"].append(filename)
            continue
        url = __save_image(
            uid=uid,
            filename=file.filename,
            file=file.file,
            file_size=file.size,
            content_type=file.content_type,
            resize_threshold=resize_threshold,
        )
        if url == "":
            res["errFiles"].append(filename)
            continue
        res["succMap"][filename] = url
    return res


def upload_bytes_to_cos(
        uid: str,
        file_size: int,
        content_type: str,
        file: BinaryIO,
        resize_threshold: int,
        filename: str,
        fid: str,
) -> str:
    token = None
    scheme = 'https'

    settings = config.get_settings()
    secret_id = settings.COS_SECRET_ID
    secret_key = settings.COS_SECRET_KEY
    region = settings.COS_REGION
    domain = None
    cos_conf = CosConfig(
        Region=region,
        SecretId=secret_id,
        SecretKey=secret_key,
        Token=token,
        Domain=domain,
        Scheme=scheme,
    )
    client = CosS3Client(cos_conf)

    settings = config.get_settings()
    key = f"userData/{uid}/{fid}"

    url = f"https://{settings.COS_BUCKET_NAME}.cos.{settings.COS_REGION}.myqcloud.com/{key}"

    doc = COLL.user_file.find_one({"uid": uid, "fid": fid})
    if doc:
        return url

    try:
        _ = client.head_object(
            Bucket=settings.COS_BUCKET_NAME,
            Key=key
        )
        return url
    except CosServiceError as e:
        if e.get_status_code() != 404:
            return url
    body = __get_out_bytes(
        file_size=file_size,
        content_type=content_type,
        file=file,
        resize_threshold=resize_threshold
    )
    # can raise error
    _ = client.put_object(
        Bucket=settings.COS_BUCKET_NAME,
        Body=body,
        Key=key,
        StorageClass='STANDARD',  # 'STANDARD'|'STANDARD_IA'|'ARCHIVE',
        EnableMD5=False,
        ContentType=content_type,
    )
    insert_new_file(
        uid=uid,
        fid=fid,
        filename=filename,
        file_size=body.tell(),
    )
    return url


def upload_to_local_storage(
        uid: str,
        file_size: int,
        content_type: str,
        file: BinaryIO,
        resize_threshold: int,
        filename: str,
        fid: str,
) -> str:
    # upload to local storage
    img_path = config.get_settings().LOCAL_STORAGE_PATH / ".data" / "files" / fid
    img_path.parent.mkdir(parents=True, exist_ok=True)
    url = f"/files/{fid}"
    if img_path.exists():
        # skip the same image
        return url
    try:
        out_bytes = __get_out_bytes(
            file_size=file_size,
            content_type=content_type,
            file=file,
            resize_threshold=resize_threshold
        )
        Image.open(out_bytes).save(img_path)
    except (
            FileNotFoundError,
            OSError,
            UnidentifiedImageError,
            ValueError,
            TypeError,
    ):
        return ""
    # point move to the end
    out_bytes.seek(0, io.SEEK_END)
    size = out_bytes.tell()
    insert_new_file(
        uid=uid,
        fid=fid,
        filename=filename,
        file_size=size,
    )
    return url


def insert_new_file(
        uid: str,
        fid: str,
        filename: str,
        file_size: int,
):
    doc: UserFile = {
        "_id": ObjectId(),
        "uid": uid,
        "fid": fid,
        "filename": filename,
        "size": file_size,
    }
    COLL.user_file.insert_one(doc)
    update_used_space(uid=uid, delta=file_size)



import unittest

from rethink.models.search.idf import TFIDF

docs = {
    "1": "The stop_words_ attribute can get large and increase the model size when pickling. This attribute is provided"
         " only for introspection and can be safely removed using delattr or set to None before pickling.",
    "2": "为(B, M, E, S): {B:begin, M:middle, E:end, S:single}。"
         "分别代表每个状态代表的是该字在词语中的位置，B代表该字是词语中的起始字，M代表是词语中的中间字，E代表是词语中的结束字，"
         "S则代表是单字成词。",
    "3": "我们在这个短片简介中提到过。 机器怎么理解句子一直是一个难题，以前有人尝试将用句子中出现的词语频率来表达这个句子的含义（TF-IDF）。"
         " 也有人想把句子中的词语先向量化，然后再叠加句子中所有向量化的词语来表达一句话。"
         " 这些在的确都是一种计算机表达句子含义的方式，"
         "但是不一定会非常准确。因为他们都只是一种对词语理解的简单加工方式，"
         "有的忽略了词语的表达顺序， 有的忽略了词语的组合模式。这往往导致计算机不能非常准确的理解句子。",
    "4": "我们在这个短片简介中提到过。 机器怎么理解句子一直是一个难题，"
         "以前有人尝试将用句子中出现的词语频率来表达这个句子的含义（TF-IDF）。",
    "5": "The hospital was built in 2011 with donations from Indonesian citizens and organisations,"
         " including the Indonesian Red Cross Society and"
         " the Muhammadiyah Society, one of Indonesia’s largest "
         "Muslim organisations. It was officially inaugurated in "
         "2016 by the then-Indonesian vice president, Jusuf Kalla. Three Indonesian volunteers with the "
         "Indonesian humanitarian organisation the Medical Emergency"
         " Rescue Committee (MER-C), which organised the "
         "donations to build the hospital, are currently based in north Gaza.",
    "6": "The hospital was built in 2011 with donations from Indonesian citizens and organisations,",

}


class TFIDFTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.tfidf = TFIDF()

    def test_tokenizer(self):
        for s, t in [
            ("计算机理解句子 apple is a fruit. ", ['计算', '算机', '计算机', '理解', '句子', 'apple', 'a', 'fruit']),
            ("wdasd dqwwq ffger 问答apple使用Python", ['wdasd', 'dqwwq', 'ffger', '问答', 'apple', '使用', 'python']),
            (
                    "The hospital was built in 2011 with donations from Indonesian citizens and organisations,",
                    ['hospital', 'was', 'built', '2011', 'donations', 'indonesian', 'citizens', 'organisations']
            ),
        ]:
            res = list(self.tfidf.tokenizer(s))
            self.assertEqual(t, res)

    def test_search(self):
        _ds = []
        nids = []
        for k, v in docs.items():
            _ds.append(v)
            nids.append(k)

        vec = self.tfidf.docs2vec(_ds)
        self.assertEqual((6, 555861), vec.shape)

        res = self.tfidf.search("计算机理解句子", docs_vec=vec, n=2)
        self.assertEqual(["3", "4"], [nids[i] for i in res])
        res = self.tfidf.search("hospital built", docs_vec=vec, n=2)
        self.assertEqual(["6", "5"], [nids[i] for i in res])



from fastapi import Request, APIRouter

from rethink.controllers import oauth as co
from rethink.controllers.schemas.user import LoginResponse
from rethink.routes.utils import measure_time_spend

router = APIRouter(
    prefix="/oauth",
    tags=["oauth"],
    responses={404: {"description": "Not found"}},
)


@router.get(path="/login/github", response_model=co.OAuthResponse)
@measure_time_spend
async def login_github() -> co.OAuthResponse:
    return await co.login_github()


@router.get(path="/callback/github", response_model=LoginResponse)
@measure_time_spend
async def callback_github(request: Request) -> LoginResponse:
    return await co.callback_github(req=request)

# @router.get(path="/login/facebook", response_model=co.OAuthResponse)
# @measure_time_spend
# async def login_facebook() -> co.OAuthResponse:
#     return await co.login_facebook()
#
#
# @router.get(path="/callback/facebook", response_model=cu.UserLoginResponse)
# @measure_time_spend
# async def callback_facebook(request: Request) -> cu.UserLoginResponse:
#     return await co.callback_facebook(req=request)



import logging

# import os
# from logging.handlers import RotatingFileHandler

# disable 3rd party log
logging.basicConfig(
    level=logging.ERROR,
    format="%(levelname)s | %(asctime)s | %(name)s | %(filename)s[line:%(lineno)d]: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger = logging.getLogger("rethink")
logger.setLevel(logging.INFO)



from typing import List, Optional

from fastapi import Depends, APIRouter, UploadFile, Request, Header
from typing_extensions import Annotated

from rethink.controllers import schemas
from rethink.controllers.auth import token2uid
from rethink.controllers.files import upload_files
from rethink.controllers.utils import TokenDecode
from rethink.routes.utils import measure_time_spend

router = APIRouter(
    prefix="/api",
    tags=["files"],
    responses={404: {"description": "Not found"}},
)


@router.post(
    path="/files/obsidian",
    response_model=schemas.files.FileUploadResponse,
)
@measure_time_spend
async def upload_obsidian_files(
        token_decode: Annotated[TokenDecode, Depends(token2uid)],
        files: List[UploadFile],
) -> schemas.files.FileUploadResponse:
    return upload_files.upload_obsidian_files(
        td=token_decode,
        files=files,
    )


@router.post(
    path="/files/text",
    response_model=schemas.files.FileUploadResponse,
)
@measure_time_spend
async def upload_obsidian_files(
        token_decode: Annotated[TokenDecode, Depends(token2uid)],
        files: List[UploadFile],
) -> schemas.files.FileUploadResponse:
    return upload_files.upload_text_files(
        td=token_decode,
        files=files,
    )


@router.get(
    path="/files/uploadProcess",
    response_model=schemas.files.FileUploadProcessResponse,
)
@measure_time_spend
async def get_upload_process(
        token_decode: Annotated[TokenDecode, Depends(token2uid)],
        rid: Optional[str] = Header(None),
) -> schemas.files.FileUploadProcessResponse:
    return upload_files.get_upload_process(
        td=token_decode,
        rid=rid,
    )


@router.post(
    path="/files/imageUploadVditor",
    response_model=schemas.files.ImageVditorUploadResponse,
)
@measure_time_spend
async def upload_image_vditor(
        token_decode: Annotated[TokenDecode, Depends(token2uid)],
        req: Request,
) -> schemas.files.ImageVditorUploadResponse:
    form = await req.form()
    file = form.get("file[]")
    return upload_files.upload_image_vditor(
        td=token_decode,
        file=file,
    )


@router.post(
    path="/files/imageFetchVditor",
    response_model=schemas.files.ImageVditorFetchResponse,
)
@measure_time_spend
async def fetch_image_vditor(
        token_decode: Annotated[TokenDecode, Depends(token2uid)],
        req: schemas.files.ImageVditorFetchRequest,
) -> schemas.files.ImageVditorFetchResponse:
    return upload_files.fetch_image_vditor(
        td=token_decode,
        req=req,
    )



from typing import Optional

from fastapi import Depends, APIRouter, Header
from typing_extensions import Annotated

from rethink.controllers import schemas
from rethink.controllers.auth import token2uid
from rethink.controllers.search import node_search
from rethink.controllers.utils import TokenDecode
from rethink.routes.utils import measure_time_spend

router = APIRouter(
    prefix="/api",
    tags=["search"],
    responses={404: {"description": "Not found"}},
)


@router.post(
    path="/search/cursor",
    response_model=schemas.search.CursorQueryResponse,
)
@measure_time_spend
async def cursor_query(
        req: schemas.search.CursorQueryRequest,
        token_decode: Annotated[TokenDecode, Depends(token2uid)]
) -> schemas.search.CursorQueryResponse:
    return node_search.cursor_query(
        td=token_decode,
        req=req,
    )


@router.post(
    path="/search/node",
    response_model=schemas.node.NodesInfoResponse,
)
@measure_time_spend
async def search_user_nodes(
        req: schemas.search.SearchUserNodesRequest,
        token_decode: Annotated[TokenDecode, Depends(token2uid)]
) -> schemas.node.NodesInfoResponse:
    return node_search.search_user_nodes(
        td=token_decode,
        req=req,
    )


@router.put(
    path="/search/cursorSelect",
    response_model=schemas.base.AcknowledgeResponse,
)
@measure_time_spend
async def add_to_recent_cursor_search(
        req: schemas.search.AddToRecentSearchHistRequest,
        token_decode: Annotated[TokenDecode, Depends(token2uid)]
) -> schemas.base.AcknowledgeResponse:
    return node_search.add_to_recent_cursor_search(
        td=token_decode,
        req=req,
    )


@router.get(
    path="/search/recent",
    response_model=schemas.search.GetRecentSearchResponse,
)
@measure_time_spend
async def get_recent_search(
        token_decode: Annotated[TokenDecode, Depends(token2uid)],
        rid: Optional[str] = Header(None),
) -> schemas.search.GetRecentSearchResponse:
    return node_search.get_recent(
        td=token_decode,
        rid=rid,
    )


@router.put(
    path="/search/recent",
    response_model=schemas.base.AcknowledgeResponse,
)
@measure_time_spend
async def put_recent_search(
        req: schemas.search.PutRecentSearchRequest,
        token_decode: Annotated[TokenDecode, Depends(token2uid)]
) -> schemas.base.AcknowledgeResponse:
    return node_search.put_recent(
        td=token_decode,
        req=req,
    )



import datetime
from dataclasses import dataclass, field

from rethink import const


@dataclass
class TokenDecode:
    code: const.Code
    uid: str = ""
    language: str = field(default=const.Language.EN.value)


def datetime2str(dt: datetime.datetime) -> str:
    return dt.strftime('%Y-%m-%dT%H:%M:%SZ')



import datetime
import re
from pathlib import Path
from typing import List, Optional, Set, Tuple

from bson import ObjectId
from bson.tz_util import utc

from rethink import config, const
from rethink.logger import logger
from . import user, tps, utils, db_ops
from .database import COLL
from .search import user_node

AT_PTN = re.compile(r'\[@[ \w\u4e00-\u9fa5！？。，￥【】「」]+?]\(([\w/]+?)\)', re.MULTILINE)
CURSOR_AT_PTN = re.compile(r'\s+?@([\w ]*)$')


def __get_linked_nodes(new_md) -> Tuple[set, const.Code]:
    # last first
    cache_current_to_nid: Set[str] = set()

    for match in list(AT_PTN.finditer(new_md))[::-1]:
        l0, l1 = match.span(1)
        link = new_md[l0:l1]
        if link.startswith("/n/"):
            # check existed nodes
            to_nid = link[3:]
            cache_current_to_nid.add(to_nid)
    return cache_current_to_nid, const.Code.OK


def __flush_to_node_ids(
        nid: str,
        orig_to_nid: List[str],
        new_md: str
) -> Tuple[List[str], const.Code]:
    new_to_nid, code = __get_linked_nodes(new_md=new_md)
    if code != const.Code.OK:
        return [], code

    # remove fromNodes for linked nodes
    orig_to_nid = set(orig_to_nid)
    for to_nid in orig_to_nid.difference(new_to_nid):
        db_ops.remove_from_node(from_nid=nid, to_nid=to_nid)

    # add fromNodes for linked nodes
    for to_nid in new_to_nid.difference(orig_to_nid):
        db_ops.node_add_to_set(id_=to_nid, key="fromNodeIds", value=nid)

    return list(new_to_nid), const.Code.OK


def __local_usage_write_file(nid: str, md: str):
    if not config.is_local_db():
        return
    md_dir = Path(config.get_settings().LOCAL_STORAGE_PATH) / ".data" / "md"
    md_dir.mkdir(parents=True, exist_ok=True)
    with open(md_dir / f"{nid}.md", "w", encoding="utf-8") as f:
        f.write(md)


def __local_usage_delete_files(nids: List[str]):
    for nid in nids:
        md_dir = Path(config.get_settings().LOCAL_STORAGE_PATH) / ".data" / "md" / f"{nid}.md"
        if md_dir.exists():
            md_dir.unlink()


def add(
        uid: str,
        md: str,
        type_: int = const.NodeType.MARKDOWN.value,
        from_nid: str = "",
) -> Tuple[Optional[tps.Node], const.Code]:
    md = md.strip()
    if len(md) > const.MD_MAX_LENGTH:
        return None, const.Code.NOTE_EXCEED_MAX_LENGTH

    u, code = user.get(uid=uid)
    if code != const.Code.OK:
        return None, code

    new_size = len(md.encode("utf-8"))
    if user.user_space_not_enough(u=u):
        return None, const.Code.USER_SPACE_NOT_ENOUGH

    title, snippet = utils.preprocess_md(md)

    nid = utils.short_uuid()

    from_nids = []
    if from_nid != "":
        from_nids.append(from_nid)
        res = db_ops.node_add_to_set(from_nid, "toNodeIds", nid)
        if res.modified_count != 1:
            return None, const.Code.OPERATION_FAILED

    new_to_node_ids = []
    if md != "":
        new_to_node_ids, code = __flush_to_node_ids(nid=nid, orig_to_nid=[], new_md=md)
        if code != const.Code.OK:
            return None, code
    _id = ObjectId()
    data: tps.Node = {
        "_id": _id,
        "id": nid,
        "uid": uid,
        "title": title,
        "snippet": snippet,
        "md": md,
        "type": type_,
        "disabled": False,
        "inTrash": False,
        "modifiedAt": _id.generation_time,
        "inTrashAt": None,
        "fromNodeIds": from_nids,
        "toNodeIds": new_to_node_ids,
        "searchKeys": utils.txt2search_keys(title),
    }
    res = COLL.nodes.insert_one(data)
    if not res.acknowledged:
        return None, const.Code.OPERATION_FAILED

    user.update_used_space(uid=uid, delta=new_size)

    __local_usage_write_file(nid=nid, md=md)
    return data, const.Code.OK


def __set_linked_nodes(
        docs: List[tps.Node],
        with_disabled: bool = False,
):
    for doc in docs:
        doc["fromNodes"] = list(
            COLL.nodes.find({
                "id": {"$in": doc["fromNodeIds"]},
                "disabled": with_disabled,
            }))
        doc["toNodes"] = list(
            COLL.nodes.find({
                "id": {"$in": doc["toNodeIds"]},
                "disabled": with_disabled,
            }))


def get(
        uid: str,
        nid: str,
        with_disabled: bool = False,
        in_trash: bool = False,
) -> Tuple[Optional[tps.Node], const.Code]:
    docs, code = get_batch(
        uid=uid,
        nids=[nid],
        with_disabled=with_disabled,
        in_trash=in_trash,
    )
    return docs[0] if len(docs) > 0 else None, code


def get_batch(
        uid: str,
        nids: List[str],
        with_disabled: bool = False,
        in_trash: bool = False,
) -> Tuple[List[tps.Node], const.Code]:
    c = {"id": {"$in": nids}, "uid": uid, "inTrash": in_trash}
    if not with_disabled:
        c["disabled"] = False
    docs = list(COLL.nodes.find(c))
    if len(docs) != len(nids):
        return [], const.Code.NODE_NOT_EXIST

    __set_linked_nodes(
        docs=docs,
        with_disabled=with_disabled,
    )
    return docs, const.Code.OK


def update(
        uid: str,
        nid: str,
        md: str,
        refresh_on_same_md: bool = False,
) -> Tuple[Optional[tps.Node], const.Code]:
    md = md.strip()
    if len(md) > const.MD_MAX_LENGTH:
        return None, const.Code.NOTE_EXCEED_MAX_LENGTH
    u, code = user.get(uid=uid)
    if code != const.Code.OK:
        return None, const.Code.ACCOUNT_OR_PASSWORD_ERROR
    if user.user_space_not_enough(u=u):
        return None, const.Code.USER_SPACE_NOT_ENOUGH

    title, snippet = utils.preprocess_md(md)

    n, code = get(uid=uid, nid=nid)
    if code != const.Code.OK:
        return None, code
    if n["md"] == md and not refresh_on_same_md:
        return n, const.Code.OK

    old_md_size = len(n["md"].encode("utf-8"))
    new_data = {
        "modifiedAt": datetime.datetime.now(tz=utc),
    }

    if n["title"] != title:
        # update it's title in fromNodes md's link
        from_nodes = COLL.nodes.find({"id": {"$in": n["fromNodeIds"]}})
        for from_node in from_nodes:
            new_md = utils.change_link_title(md=from_node["md"], nid=nid, new_title=title)
            n, code = update(uid=uid, nid=from_node["id"], md=new_md)
            if code != const.Code.OK:
                logger.info(f"update fromNode {from_node['id']} failed")
        new_data["title"] = title
        new_data["searchKeys"] = utils.txt2search_keys(title)

    if n["md"] != md:
        new_data["md"] = md
        if n["snippet"] != snippet:
            new_data["snippet"] = snippet

    new_data["toNodeIds"], code = __flush_to_node_ids(
        nid=n["id"], orig_to_nid=n["toNodeIds"], new_md=md)
    if code != const.Code.OK:
        return None, code

    if not config.is_local_db():
        doc = COLL.nodes.find_one_and_update(
            {"id": nid},
            {"$set": new_data},
            return_document=True,  # return updated doc
        )
    else:
        # local db not support find_one_and_update
        res = COLL.nodes.update_one(
            {"id": nid},
            {"$set": new_data}
        )
        if res.modified_count != 1:
            logger.error(f"update node {nid} failed")
            return None, const.Code.OPERATION_FAILED
        doc = COLL.nodes.find_one({"id": nid})

    if doc is None:
        return None, const.Code.NODE_NOT_EXIST
    __set_linked_nodes(
        docs=[doc],
        with_disabled=False,
    )
    user.update_used_space(uid=uid, delta=len(md.encode("utf-8")) - old_md_size)

    __local_usage_write_file(nid=nid, md=md)
    return doc, const.Code.OK


def cursor_query(
        uid: str,
        nid: str,
        cursor_text: str,
) -> List[tps.Node]:
    # if cursor_text.startswith("@"):
    #     query = cursor_text[1:].strip()
    # else:
    #     found = CURSOR_AT_PTN.search(cursor_text)
    #     if found is None:
    #         return None, []
    #     query = found.group(1).strip()

    query = cursor_text.strip()

    if query == "":
        u = COLL.users.find_one({"id": uid})
        if u is None:
            return []
        rn = u["lastState"]["recentCursorSearchSelectedNIds"]
        try:
            rn.remove(nid)
        except ValueError:
            pass
        return sorted(list(COLL.nodes.find({"id": {"$in": rn}})), key=lambda x: rn.index(x["id"]))
    nodes_found, _ = user_node(
        uid=uid,
        query=query,
        sort_key="modifiedAt",
        sort_order=-1,
        page=0,
        page_size=8,
        nid_exclude=[nid],
    )
    return nodes_found


def to_trash(uid: str, nid: str) -> const.Code:
    return batch_to_trash(uid=uid, nids=[nid])


def batch_to_trash(uid: str, nids: List[str]) -> const.Code:
    ns, code = get_batch(uid=uid, nids=nids, with_disabled=True, in_trash=False)
    if code != const.Code.OK:
        return code
    u = COLL.users.find_one({"id": uid})
    changed = False
    for n in ns:
        try:
            u["lastState"]["recentCursorSearchSelectedNIds"].remove(n["id"])
            changed = True
        except ValueError:
            pass
    if changed:
        res = COLL.users.update_one({"id": uid}, {"$set": {
            "lastState.recentCursorSearchSelectedNIds": u["lastState"]["recentCursorSearchSelectedNIds"]
        }})
        if res.modified_count != 1:
            logger.error(f"update user {uid} failed")
            return const.Code.OPERATION_FAILED
    res = COLL.nodes.update_many({"id": {"$in": nids}}, {"$set": {
        "inTrash": True,
        "inTrashAt": datetime.datetime.now(tz=utc)
    }})
    if res.modified_count != len(nids):
        logger.error(f"update nodes {nids} failed")
        return const.Code.OPERATION_FAILED
    return const.Code.OK


def get_nodes_in_trash(uid: str, page: int, page_size: int) -> Tuple[List[tps.Node], int]:
    condition = {
        "uid": uid,
        "disabled": False,
        "inTrash": True,
    }
    docs = COLL.nodes.find(condition).sort([("inTrashAt", -1), ("_id", -1)])
    total = COLL.nodes.count_documents(condition)
    if page_size > 0:
        docs = docs.skip(page * page_size).limit(page_size)

    return list(docs), total


def restore_from_trash(uid: str, nid: str) -> const.Code:
    return restore_batch_from_trash(uid=uid, nids=[nid])


def restore_batch_from_trash(uid: str, nids: List[str]) -> const.Code:
    ns, code = get_batch(uid=uid, nids=nids, with_disabled=False, in_trash=True)
    if code != const.Code.OK:
        return code

    # remove nodes
    res = COLL.nodes.update_many({"id": {"$in": nids}}, {"$set": {
        "inTrash": False,
        "inTrashAt": None,
    }})
    if res.modified_count != len(nids):
        logger.error(f"restore nodes {nids} failed")
        return const.Code.OPERATION_FAILED
    return const.Code.OK


def delete(uid: str, nid: str) -> const.Code:
    return batch_delete(uid=uid, nids=[nid])


def batch_delete(uid: str, nids: List[str]) -> const.Code:
    ns, code = get_batch(uid=uid, nids=nids, with_disabled=True, in_trash=True)
    if code != const.Code.OK:
        return code

    # remove fromNodes for linked nodes, not necessary
    # for n in ns:
    #     for linked_nid in n["fromNodeIds"]:
    #         db_ops.remove_to_node(from_nid=linked_nid, to_nid=n["id"])

    used_space_delta = 0
    # remove toNodes for linked nodes
    for n in ns:
        used_space_delta -= len(n["md"].encode("utf-8"))
        for linked_nid in n["toNodeIds"]:
            db_ops.remove_from_node(from_nid=n["id"], to_nid=linked_nid)

    # delete user file
    user.update_used_space(uid=uid, delta=used_space_delta)

    # remove node
    res = COLL.nodes.delete_many({"id": {"$in": nids}, "uid": uid})
    if res.deleted_count != len(nids):
        logger.error(f"delete nodes {nids} failed")
        return const.Code.OPERATION_FAILED

    __local_usage_delete_files(nids=nids)
    return const.Code.OK


def disable(
        uid: str,
        nid: str,
) -> const.Code:
    if not user.is_exist(uid=uid):
        return const.Code.ACCOUNT_OR_PASSWORD_ERROR
    res = COLL.nodes.update_one(
        {"id": nid},
        {"$set": {"disabled": True}}
    )
    if res.modified_count != 1:
        logger.error(f"disable node {nid} failed")
        return const.Code.OPERATION_FAILED
    return const.Code.OK


def new_user_add_default_nodes(language: str, uid: str) -> const.Code:
    lns = const.NEW_USER_DEFAULT_NODES[language]
    n, code = add(
        uid=uid,
        md=lns[0],
    )
    if code != const.Code.OK:
        return code
    _, code = add(
        uid=uid,
        md=lns[1].format(n["id"]),
    )
    return code



from typing import List

from fastapi import UploadFile

from rethink import const, models
from rethink.controllers import schemas
from rethink.controllers.utils import TokenDecode
from ..utils import datetime2str


def upload_obsidian_files(
        td: TokenDecode,
        files: List[UploadFile]
) -> schemas.files.FileUploadResponse:
    if td.code != const.Code.OK:
        return schemas.files.FileUploadResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId="",
            failedFilename="",
        )
    code = models.files.upload_obsidian(uid=td.uid, zipped_files=files)

    return schemas.files.FileUploadResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId="",
    )


def upload_text_files(
        td: TokenDecode,
        files: List[UploadFile]
) -> schemas.files.FileUploadResponse:
    if td.code != const.Code.OK:
        return schemas.files.FileUploadResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId="",
            failedFilename="",
        )
    code = models.files.upload_text(uid=td.uid, files=files)
    return schemas.files.FileUploadResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId="",
        failedFilename="",
    )


def get_upload_process(
        td: TokenDecode,
        rid: str,
) -> schemas.files.FileUploadProcessResponse:
    if td.code != const.Code.OK:
        return schemas.files.FileUploadProcessResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId=rid,
            process=0.,
            type="",
            startAt="",
            running=False,
            problemFiles=[],
        )
    doc = models.files.get_upload_process(uid=td.uid)
    if doc is None:
        return schemas.files.FileUploadProcessResponse(
            code=const.Code.OK.value,
            message=const.get_msg_by_code(const.Code.OK, td.language),
            requestId=rid,
            process=0.,
            type="",
            startAt="",
            running=False,
            problemFiles=[],
        )
    code = const.INT_CODE_MAP[doc["code"]]
    return schemas.files.FileUploadProcessResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId=rid,
        process=doc["process"],
        type=doc["type"],
        startAt=datetime2str(doc["startAt"]),
        running=doc["running"],
        problemFiles=doc["problemFiles"],
    )


def upload_image_vditor(
        td: TokenDecode,
        file: UploadFile,
) -> schemas.files.ImageVditorUploadResponse:
    if td.code != const.Code.OK:
        return schemas.files.ImageVditorUploadResponse(
            code=td.code.value,
            msg=const.get_msg_by_code(td.code, td.language),
            data={},
        )
    res = models.files.upload_image_vditor(uid=td.uid, files=[file])
    return schemas.files.ImageVditorUploadResponse(
        code=const.Code.OK.value,
        msg=const.get_msg_by_code(const.Code.OK, td.language),
        data=res,
    )


def fetch_image_vditor(
        td: TokenDecode,
        req: schemas.files.ImageVditorFetchRequest,
) -> schemas.files.ImageVditorFetchResponse:
    if td.code != const.Code.OK:
        return schemas.files.ImageVditorFetchResponse(
            code=td.code.value,
            msg=const.get_msg_by_code(td.code, td.language),
            data={},
        )
    new_url, code = models.files.fetch_image_vditor(uid=td.uid, url=req.url)
    return schemas.files.ImageVditorFetchResponse(
        code=const.Code.OK.value,
        msg=const.get_msg_by_code(const.Code.OK, td.language),
        data=schemas.files.ImageVditorFetchResponse.Data(
            originalURL=req.url,
            url=new_url,
        ),
    )



import datetime
from typing import List, Optional, Dict

from bson import ObjectId
from typing_extensions import TypedDict


class LinkedNode(TypedDict):
    _id: ObjectId
    id: str
    md: str
    title: str
    type: int  # const.NodeType.MARKDOWN.value
    disabled: bool
    modifiedAt: datetime.datetime


class Node(TypedDict):
    _id: ObjectId
    id: str
    uid: str
    md: str
    title: str
    snippet: str
    type: int  # const.NodeType.MARKDOWN.value
    searchKeys: str
    disabled: bool
    inTrash: bool
    modifiedAt: datetime.datetime
    inTrashAt: Optional[datetime.datetime]
    fromNodeIds: List[str]
    toNodeIds: List[str]
    fromNodes: Optional[List[LinkedNode]]
    toNodes: Optional[List[LinkedNode]]


class _LastState(TypedDict):
    nodeDisplayMethod: int
    nodeDisplaySortKey: str
    recentSearch: List[str]
    recentCursorSearchSelectedNIds: List[str]


class UserMeta(TypedDict):
    _id: ObjectId
    id: str
    source: int
    account: str
    nickname: str
    email: str
    avatar: str
    hashed: str
    disabled: bool
    modifiedAt: datetime.datetime
    language: str
    usedSpace: int
    type: int
    lastState: _LastState


class UserFile(TypedDict):
    _id: ObjectId
    uid: str
    fid: str
    filename: str
    size: int


class ImportData(TypedDict):
    _id: ObjectId
    uid: str
    process: int
    type: str
    startAt: datetime.datetime
    running: bool
    problemFiles: List[str]
    code: int
    obsidian: Dict[str, str]  # filename: nid



import io
import os
import unittest
from textwrap import dedent
from zipfile import ZipFile, BadZipFile

from bson import ObjectId

from rethink.models.files import file_ops


class UnzipTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.orig_folder_data = {
            os.path.join("a", "test.txt"): b"hello world",
            os.path.join("a", "a.txt"): b"a",
            os.path.join("a", "v", "q文档.txt"): "阿斯顿请问".encode("utf-8"),
        }
        cls.orig_data = {
            "test.txt": b"hello world",
            "a.txt": b"a",
            "q文档.txt": "阿斯顿请问".encode("utf-8"),
        }

    def test_unzip_folder(self):
        zip_bytes = ZipFile("test.zip", "w")
        for k, v in self.orig_folder_data.items():
            zip_bytes.writestr(k, v)
        zip_bytes.close()

        # unzip
        with open("test.zip", "rb") as f:
            extracted_files = file_ops.unzip_file(f.read())
        for filename, data in extracted_files.items():
            self.assertEqual(
                self.orig_folder_data[os.path.join("a", *filename.split("/"))],
                data["file"],
                msg=str(self.orig_folder_data))
        os.remove("test.zip")

    def test_unzip_files(self):
        zip_bytes = ZipFile("test.zip", "w")
        for k, v in self.orig_data.items():
            zip_bytes.writestr(k, v)
        zip_bytes.close()

        # unzip
        with open("test.zip", "rb") as f:
            extracted_files = file_ops.unzip_file(f.read())
        for filename, data in extracted_files.items():
            self.assertEqual(self.orig_data[filename], data["file"])
        os.remove("test.zip")

    def test_unzip_not_zip(self):
        with open("test.zip", "wb") as f:
            f.write(b"hello world")
        with open("test.zip", "rb") as f:
            with self.assertRaises(BadZipFile):
                _ = file_ops.unzip_file(f.read())
        os.remove("test.zip")

    def test_file_hash(self):
        self.assertEqual(
            "d41d8cd98f00b204e9800998ecf8427e",
            file_ops.file_hash(io.BytesIO(b"")))
        bio = io.BytesIO(b"The quick brown fox jumps over the lazy dog")
        self.assertEqual(
            "9e107d9d372bb6826bd81d3542a419d6",
            file_ops.file_hash(bio))

        self.assertEqual(b"The quick brown fox jumps over the lazy dog", bio.read())

    def test_replace_inner_link(self):
        md = dedent("""\
            # 123
            ddd qwd [[123]] 345
            [[我哦]]
            """)
        o1 = str(ObjectId())
        o2 = str(ObjectId())
        filename2nid = {"123": o1, "我哦": o2}
        res = file_ops.replace_inner_link_and_upload_image(
            uid="",
            md=md,
            exist_filename2nid=filename2nid,
            img_path_dict={},
            img_name_dict={},
            resize_threshold=0,
        )
        self.assertEqual(dedent(f"""\
            # 123
            ddd qwd [@123](/n/{o1}) 345
            [@我哦](/n/{o2})
            """), res)
        self.assertEqual({"123": o1, "我哦": o2}, filename2nid)

    def test_replace_inner_link_new(self):
        md = dedent("""\
            # 123
            ddd qwd [[123]] 345
            [[我哦]]
            """)

        o1 = str(ObjectId())
        filename2nid = {"123": o1}
        res = file_ops.replace_inner_link_and_upload_image(
            uid="",
            md=md,
            exist_filename2nid=filename2nid,
            img_path_dict={},
            img_name_dict={},
            resize_threshold=0,
        )
        self.assertEqual(dedent(f"""\
            # 123
            ddd qwd [@123](/n/{o1}) 345
            [@我哦](/n/{filename2nid['我哦']})
            """), res)
        self.assertIn("我哦", filename2nid)
